# License: 3-clause BSD
# Copyright (c) 2016-2018, Ml4AAD Group (http://www.ml4aad.org/)

from inspect import signature, Signature
import math
from typing import Optional, Union, Tuple, List, Callable, Dict, Any

import numpy as np
import sklearn.gaussian_process.kernels
import scipy.optimize
import scipy.spatial.distance
import scipy.special

from xbbo.surrogate.gp_prior import Prior
import skopt.learning.gaussian_process.kernels as kernels


# This file contains almost no type annotations to simplify comparing it to the original scikit-learn version!


def get_conditional_hyperparameters(X: np.ndarray, Y: Optional[np.ndarray]) -> np.ndarray:
    # Taking care of conditional hyperparameters according to Levesque et al.
    X_cond = X <= -1
    if Y is not None:
        Y_cond = Y <= -1
    else:
        Y_cond = X <= -1
    active = ~((np.expand_dims(X_cond, axis=1) != Y_cond).any(axis=2))
    return active


class MagicMixin:
    # This is a mixin for a kernel to override functions of the kernel. Because it overrides functions of the kernel,
    # it needs to be placed first in the inheritance hierarchy. For this reason it is not possible to subclass the
    # Mixin from the kernel class because this will prevent it from being instantiatable. Therefore, mypy won't know
    # about anything related to the superclass and I had to add a few type:ignore statements when accessing a member
    # that is declared in the superclass such as self.has_conditions, self._call, super().get_params etc.

    prior = None  # type: Optional[Prior]

    def __call__(
            self,
            X: np.ndarray,
            Y: Optional[np.ndarray] = None,
            eval_gradient: bool = False,
            active: Optional[np.ndarray] = None,
    ) -> Union[np.ndarray, Tuple[np.ndarray, np.ndarray]]:

        if active is None and self.has_conditions:  # type: ignore[attr-defined] # noqa F821
            if self.operate_on is None:
                active = get_conditional_hyperparameters(X, Y)
            else:
                if Y is None:
                    active = get_conditional_hyperparameters(X[:, self.operate_on], None)
                else:
                    active = get_conditional_hyperparameters(X[:, self.operate_on], Y[:, self.operate_on])

        if self.operate_on is None:
            rval = self._call(X, Y, eval_gradient, active)  # type: ignore[attr-defined] # noqa F821
        else:
            if Y is None:
                rval = self._call(  # type: ignore[attr-defined] # noqa F821
                    X=X[:, self.operate_on].reshape([-1, self.len_active]),
                    Y=None,
                    eval_gradient=eval_gradient,
                    active=active,
                )
                X = X[:, self.operate_on].reshape((-1, self.len_active))
            else:
                rval = self._call(  # type: ignore[attr-defined] # noqa F821
                    X=X[:, self.operate_on].reshape([-1, self.len_active]),
                    Y=Y[:, self.operate_on].reshape([-1, self.len_active]),
                    eval_gradient=eval_gradient,
                    active=active,
                )
                X = X[:, self.operate_on].reshape((-1, self.len_active))
                Y = Y[:, self.operate_on].reshape((-1, self.len_active))

        return rval

    def __add__(self, b: Union[kernels.Kernel, float]) -> kernels.Sum:
        if not isinstance(b, kernels.Kernel):
            return Sum(self, ConstantKernel(b))
        return Sum(self, b)

    def __radd__(self, b: Union[kernels.Kernel, float]) -> kernels.Sum:
        if not isinstance(b, kernels.Kernel):
            return Sum(ConstantKernel(b), self)
        return Sum(b, self)

    def __mul__(self, b: Union[kernels.Kernel, float]) -> kernels.Product:
        if not isinstance(b, kernels.Kernel):
            return Product(self, ConstantKernel(b))
        return Product(self, b)

    def __rmul__(self, b: Union[kernels.Kernel, float]) -> kernels.Product:
        if not isinstance(b, kernels.Kernel):
            return Product(ConstantKernel(b), self)
        return Product(b, self)

    def _signature(self, func: Callable) -> Signature:
        try:
            sig_ = self._signature_cache.get(func)  # type: Optional[Signature]
        except AttributeError:
            self._signature_cache = {}  # type: Dict[Callable, Signature]
            sig_ = None
        if sig_ is None:
            sig = signature(func)
            self._signature_cache[func] = sig
            return sig
        else:
            return sig_

    def get_params(self, deep: bool = True) -> Dict[str, Any]:
        """Get parameters of this kernel.
        Parameters
        ----------
        deep : boolean, optional
            If True, will return the parameters for this estimator and
            contained subobjects that are estimators.
        Returns
        -------
        params : mapping of string to any
            Parameter names mapped to their values.
        """
        params = dict()

        try:
            args = self._args_cache
        except AttributeError:
            # ignore[misc] looks like it catches all kinds of errors, but misc is actually a category from mypy:
            # https://mypy.readthedocs.io/en/latest/error_code_list.html#miscellaneous-checks-misc
            tmp = super().get_params(deep)  # type: ignore[misc] # noqa F821
            args = list(tmp.keys())
            # Sum and Product do not clone the 'has_conditions' attribute by default. Instead of changing their
            # get_params() method, we simply add the attribute here!
            if 'has_conditions' not in args:
                args.append('has_conditions')
            self._args_cache = args  # type: List[Union[str, Any]]

        for arg in args:
            params[arg] = getattr(self, arg, None)
        return params

    @property
    def hyperparameters(self) -> List[kernels.Hyperparameter]:
        """Returns a list of all hyperparameter specifications."""
        try:
            return self._hyperparameters_cache
        except AttributeError:
            pass

        r = super().hyperparameters  # type: ignore[misc] # noqa F821
        self._hyperparameters_cache = r  # type: List[kernels.Hyperparameter]

        return r

    @property
    def n_dims(self) -> int:
        """Returns the number of non-fixed hyperparameters of the kernel."""

        try:
            return self._n_dims_cache
        except AttributeError:
            pass

        self._n_dims_cache = -1  # type: int # I cannot use `varname: type = value` syntax because that's >=Python3.6
        self._n_dims_cache = super().n_dims  # type: ignore[misc] # noqa F821
        return self._n_dims_cache

    def clone_with_theta(self, theta: np.ndarray) -> kernels.Kernel:
        """Returns a clone of self with given hyperparameters theta.
        Parameters
        ----------
        theta : array, shape (n_dims,)
            The hyperparameters
        """
        self.theta = theta
        return self

    def set_active_dims(self, operate_on: Optional[np.ndarray] = None) -> None:
        """Sets dimensions this kernel should work on
        Parameters
        ----------
        operate_on : None, list or array, shape (n_dims,)
        """
        if operate_on is not None and type(operate_on) in (list, np.ndarray):
            if not isinstance(operate_on, np.ndarray):
                raise TypeError('argument operate_on needs to be of type np.ndarray, but is %s' % type(operate_on))
            if operate_on.dtype != np.int:
                raise ValueError('dtype of argument operate_on needs to be np.int, but is %s' % operate_on.dtype)
            self.operate_on = operate_on  # type: Optional[np.ndarray]
            self.len_active = len(operate_on)  # type: Optional[int]
        else:
            self.operate_on = None
            self.len_active = None


class Sum(MagicMixin, kernels.Sum):

    def __init__(
            self,
            k1: kernels.Kernel,
            k2: kernels.Kernel,
            operate_on: np.ndarray = None,
            has_conditions: bool = False,
    ) -> None:
        super(Sum, self).__init__(k1=k1, k2=k2)
        self.set_active_dims(operate_on)
        self.has_conditions = has_conditions

    def _call(
            self,
            X: np.ndarray,
            Y: Optional[np.ndarray] = None,
            eval_gradient: bool = False,
            active: np.ndarray = None,
    ) -> Union[np.ndarray, Tuple[np.ndarray, np.ndarray]]:
        """Return the kernel k(X, Y) and optionally its gradient.
        Parameters
        ----------
        X : array, shape (n_samples_X, n_features)
            Left argument of the returned kernel k(X, Y)
        Y : array, shape (n_samples_Y, n_features), (optional, default=None)
            Right argument of the returned kernel k(X, Y). If None, k(X, X)
            if evaluated instead.
        eval_gradient : bool (optional, default=False)
            Determines whether the gradient with respect to the kernel
            hyperparameter is determined.
        active : np.ndarray (n_samples_X, n_features) (optional)
            Boolean array specifying which hyperparameters are active.
        Returns
        -------
        K : array, shape (n_samples_X, n_samples_Y)
            Kernel k(X, Y)
        K_gradient : array (opt.), shape (n_samples_X, n_samples_X, n_dims)
            The gradient of the kernel k(X, X) with respect to the
            hyperparameter of the kernel. Only returned when eval_gradient
            is True.
        """
        if eval_gradient:
            K1, K1_gradient = self.k1(X, Y, eval_gradient=True, active=active)
            K2, K2_gradient = self.k2(X, Y, eval_gradient=True, active=active)
            return K1 + K2, np.dstack((K1_gradient, K2_gradient))
        else:
            return self.k1(X, Y, active=active) + self.k2(X, Y, active=active)


class Product(MagicMixin, kernels.Product):

    def __init__(
            self,
            k1: kernels.Kernel,
            k2: kernels.Kernel,
            operate_on: np.ndarray = None,
            has_conditions: bool = False,
    ) -> None:
        super(Product, self).__init__(k1=k1, k2=k2)
        self.set_active_dims(operate_on)
        self.has_conditions = has_conditions

    def _call(
            self,
            X: np.ndarray,
            Y: Optional[np.ndarray] = None,
            eval_gradient: bool = False,
            active: np.ndarray = None,
    ) -> Union[np.ndarray, Tuple[np.ndarray, np.ndarray]]:
        """Return the kernel k(X, Y) and optionally its gradient.
        Parameters
        ----------
        X : array, shape (n_samples_X, n_features)
            Left argument of the returned kernel k(X, Y)
        Y : array, shape (n_samples_Y, n_features), (optional, default=None)
            Right argument of the returned kernel k(X, Y). If None, k(X, X)
            if evaluated instead.
        eval_gradient : bool (optional, default=False)
            Determines whether the gradient with respect to the kernel
            hyperparameter is determined.
        active : np.ndarray (n_samples_X, n_features) (optional)
            Boolean array specifying which hyperparameters are active.
        Returns
        -------
        K : array, shape (n_samples_X, n_samples_Y)
            Kernel k(X, Y)
        K_gradient : array (opt.), shape (n_samples_X, n_samples_X, n_dims)
            The gradient of the kernel k(X, X) with respect to the
            hyperparameter of the kernel. Only returned when eval_gradient
            is True.
        """
        if eval_gradient:
            K1, K1_gradient = self.k1(X, Y, eval_gradient=True, active=active)
            K2, K2_gradient = self.k2(X, Y, eval_gradient=True, active=active)
            return K1 * K2, np.dstack((K1_gradient * K2[:, :, np.newaxis],
                                       K2_gradient * K1[:, :, np.newaxis]))
        else:
            return self.k1(X, Y, active=active) * self.k2(X, Y, active=active)


class ConstantKernel(MagicMixin, kernels.ConstantKernel):

    def __init__(
            self,
            constant_value: float = 1.0,
            constant_value_bounds: Tuple[float, float] = (1e-5, 1e5),
            operate_on: Optional[np.ndarray] = None,
            prior: Optional[Prior] = None,
            has_conditions: bool = False,
    ) -> None:

        super(ConstantKernel, self).__init__(constant_value=constant_value, constant_value_bounds=constant_value_bounds)
        self.set_active_dims(operate_on)
        self.prior = prior
        self.has_conditions = has_conditions

    def _call(
            self,
            X: np.ndarray,
            Y: Optional[np.ndarray] = None,
            eval_gradient: bool = False,
            active: Optional[np.ndarray] = None,
    ) -> Union[np.ndarray, Tuple[np.ndarray, np.ndarray]]:
        """Return the kernel k(X, Y) and optionally its gradient.
        Parameters
        ----------
        X : array, shape (n_samples_X, n_features)
            Left argument of the returned kernel k(X, Y)
        Y : array, shape (n_samples_Y, n_features), (optional, default=None)
            Right argument of the returned kernel k(X, Y). If None, k(X, X)
            if evaluated instead.
        eval_gradient : bool (optional, default=False)
            Determines whether the gradient with respect to the kernel
            hyperparameter is determined. Only supported when Y is None.
        active : np.ndarray (n_samples_X, n_features) (optional)
            Boolean array specifying which hyperparameters are active.
        Returns
        -------
        K : array, shape (n_samples_X, n_samples_Y)
            Kernel k(X, Y)
        K_gradient : array (opt.), shape (n_samples_X, n_samples_X, n_dims)
            The gradient of the kernel k(X, X) with respect to the
            hyperparameter of the kernel. Only returned when eval_gradient
            is True.
        """
        X = np.atleast_2d(X)
        if Y is None:
            Y = X
        elif eval_gradient:
            raise ValueError("Gradient can only be evaluated when Y is None.")

        K = np.full((X.shape[0], Y.shape[0]), self.constant_value,
                    dtype=np.array(self.constant_value).dtype)
        if eval_gradient:
            if not self.hyperparameter_constant_value.fixed:
                return (K, np.full((X.shape[0], X.shape[0], 1),
                                   self.constant_value,
                                   dtype=np.array(self.constant_value).dtype))
            else:
                return K, np.empty((X.shape[0], X.shape[0], 0))
        else:
            return K


class Matern(MagicMixin, kernels.Matern):

    def __init__(
            self,
            length_scale: Union[float, Tuple[float, ...]] = 1.0,
            length_scale_bounds: Union[Tuple[float, float], List[Tuple[float, float]]] = (1e-5, 1e5),
            nu: float = 1.5,
            operate_on: Optional[np.ndarray] = None,
            prior: Optional[Prior] = None,
            has_conditions: bool = False,
    ) -> None:

        super(Matern, self).__init__(length_scale=length_scale, length_scale_bounds=length_scale_bounds, nu=nu)
        self.set_active_dims(operate_on)
        self.prior = prior
        self.has_conditions = has_conditions

    def _call(
            self,
            X: np.ndarray,
            Y: Optional[np.ndarray] = None,
            eval_gradient: bool = False,
            active: Optional[np.ndarray] = None,
    ) -> Union[np.ndarray, Tuple[np.ndarray, np.ndarray]]:
        """Return the kernel k(X, Y) and optionally its gradient.
        Parameters
        ----------
        X : array, shape (n_samples_X, n_features)
            Left argument of the returned kernel k(X, Y)
        Y : array, shape (n_samples_Y, n_features), (optional, default=None)
            Right argument of the returned kernel k(X, Y). If None, k(X, X)
            if evaluated instead.
        eval_gradient : bool (optional, default=False)
            Determines whether the gradient with respect to the kernel
            hyperparameter is determined. Only supported when Y is None.
        active : np.ndarray (n_samples_X, n_features) (optional)
            Boolean array specifying which hyperparameters are active.
        Returns
        -------
        K : array, shape (n_samples_X, n_samples_Y)
            Kernel k(X, Y)
        K_gradient : array (opt.), shape (n_samples_X, n_samples_X, n_dims)
            The gradient of the kernel k(X, X) with respect to the
            hyperparameter of the kernel. Only returned when eval_gradient
            is True.
        """

        X = np.atleast_2d(X)
        length_scale = sklearn.gaussian_process.kernels._check_length_scale(X, self.length_scale)

        if Y is None:
            dists = scipy.spatial.distance.pdist(X / length_scale, metric='euclidean')
        else:
            if eval_gradient:
                raise ValueError(
                    "Gradient can only be evaluated when Y is None.")
            dists = scipy.spatial.distance.cdist(X / length_scale, Y / length_scale, metric='euclidean')

        if self.nu == 0.5:
            K = np.exp(-dists)
        elif self.nu == 1.5:
            K = dists * math.sqrt(3)
            K = (1. + K) * np.exp(-K)
        elif self.nu == 2.5:
            K = dists * math.sqrt(5)
            K = (1. + K + K ** 2 / 3.0) * np.exp(-K)
        else:  # general case; expensive to evaluate
            K = dists
            K[K == 0.0] += np.finfo(float).eps  # strict zeros result in nan
            tmp = (math.sqrt(2 * self.nu) * K)
            K.fill((2 ** (1. - self.nu)) / scipy.special.gamma(self.nu))
            K *= tmp ** self.nu
            K *= scipy.special.kv(self.nu, tmp)

        if Y is None:
            # convert from upper-triangular matrix to square matrix
            K = scipy.spatial.distance.squareform(K)
            np.fill_diagonal(K, 1)

        if active is not None:
            K = K * active

        if eval_gradient:
            if self.hyperparameter_length_scale.fixed:
                # Hyperparameter l kept fixed
                K_gradient = np.empty((X.shape[0], X.shape[0], 0))
                return K, K_gradient

            # We need to recompute the pairwise dimension-wise distances
            if self.anisotropic:
                D = (X[:, np.newaxis, :] - X[np.newaxis, :, :]) ** 2 / (length_scale ** 2)
            else:
                D = scipy.spatial.distance.squareform(dists ** 2)[:, :, np.newaxis]

            if self.nu == 0.5:
                K_gradient = K[..., np.newaxis] * D / np.sqrt(D.sum(2))[:, :, np.newaxis]
                K_gradient[~np.isfinite(K_gradient)] = 0
            elif self.nu == 1.5:
                K_gradient = 3 * D * np.exp(-np.sqrt(3 * D.sum(-1)))[..., np.newaxis]
            elif self.nu == 2.5:
                tmp = np.sqrt(5 * D.sum(-1))[..., np.newaxis]
                K_gradient = 5.0 / 3.0 * D * (tmp + 1) * np.exp(-tmp)
            else:
                # original sklearn code would approximate gradient numerically, but this would violate our assumption
                # that the kernel hyperparameters are not changed within __call__
                raise ValueError(self.nu)

            if not self.anisotropic:
                return K, K_gradient[:, :].sum(-1)[:, :, np.newaxis]
            else:
                return K, K_gradient
        else:
            return K


class RBF(MagicMixin, kernels.RBF):

    def __init__(
            self,
            length_scale: Union[float, Tuple[float, ...]] = 1.0,
            length_scale_bounds: Union[Tuple[float, float], List[Tuple[float, float]]] = (1e-5, 1e5),
            operate_on: Optional[np.ndarray] = None,
            prior: Optional[Prior] = None,
            has_conditions: bool = False,
    ) -> None:

        super(RBF, self).__init__(length_scale=length_scale, length_scale_bounds=length_scale_bounds)
        self.set_active_dims(operate_on)
        self.prior = prior
        self.has_conditions = has_conditions

    def _call(
            self,
            X: np.ndarray,
            Y: Optional[np.ndarray] = None,
            eval_gradient: bool = False,
            active: Optional[np.ndarray] = None,
    ) -> Union[np.ndarray, Tuple[np.ndarray, np.ndarray]]:

        """Return the kernel k(X, Y) and optionally its gradient.
        Parameters
        ----------
        X : array, shape (n_samples_X, n_features)
            Left argument of the returned kernel k(X, Y)
        Y : array, shape (n_samples_Y, n_features), (optional, default=None)
            Right argument of the returned kernel k(X, Y). If None, k(X, X)
            if evaluated instead.
        eval_gradient : bool (optional, default=False)
            Determines whether the gradient with respect to the kernel
            hyperparameter is determined. Only supported when Y is None.
        active : np.ndarray (n_samples_X, n_features) (optional)
            Boolean array specifying which hyperparameters are active.
        Returns
        -------
        K : array, shape (n_samples_X, n_samples_Y)
            Kernel k(X, Y)
        K_gradient : array (opt.), shape (n_samples_X, n_samples_X, n_dims)
            The gradient of the kernel k(X, X) with respect to the
            hyperparameter of the kernel. Only returned when eval_gradient
            is True.
        """

        X = np.atleast_2d(X)
        length_scale = sklearn.gaussian_process.kernels._check_length_scale(X, self.length_scale)

        if Y is None:
            dists = scipy.spatial.distance.pdist(X / length_scale, metric='sqeuclidean')
            K = np.exp(-.5 * dists)
            # convert from upper-triangular matrix to square matrix
            K = scipy.spatial.distance.squareform(K)
            np.fill_diagonal(K, 1)
        else:
            if eval_gradient:
                raise ValueError(
                    "Gradient can only be evaluated when Y is None.")
            dists = scipy.spatial.distance.cdist(X / length_scale, Y / length_scale, metric='sqeuclidean')
            K = np.exp(-.5 * dists)

        if active is not None:
            K = K * active

        if eval_gradient:
            if self.hyperparameter_length_scale.fixed:
                # Hyperparameter l kept fixed
                return K, np.empty((X.shape[0], X.shape[0], 0))
            elif not self.anisotropic or length_scale.shape[0] == 1:
                K_gradient = (K * scipy.spatial.distance.squareform(dists))[:, :, np.newaxis]
                return K, K_gradient
            elif self.anisotropic:
                # We need to recompute the pairwise dimension-wise distances
                K_gradient = (X[:, np.newaxis, :] - X[np.newaxis, :, :]) ** 2 / (length_scale ** 2)
                K_gradient *= K[..., np.newaxis]
                return K, K_gradient

        return K


class WhiteKernel(MagicMixin, kernels.WhiteKernel):

    def __init__(
            self,
            noise_level: Union[float, Tuple[float, ...]] = 1.0,
            noise_level_bounds: Union[Tuple[float, float], List[Tuple[float, float]]] = (1e-5, 1e5),
            operate_on: Optional[np.ndarray] = None,
            prior: Optional[Prior] = None,
            has_conditions: bool = False,
    ) -> None:

        super(WhiteKernel, self).__init__(noise_level=noise_level, noise_level_bounds=noise_level_bounds)
        self.set_active_dims(operate_on)
        self.prior = prior
        self.has_conditions = has_conditions

    def _call(
            self,
            X: np.ndarray,
            Y: Optional[np.ndarray] = None,
            eval_gradient: bool = False,
            active: Optional[np.ndarray] = None,
    ) -> Union[np.ndarray, Tuple[np.ndarray, np.ndarray]]:
        """Return the kernel k(X, Y) and optionally its gradient.
        Parameters
        ----------
        X : array, shape (n_samples_X, n_features)
            Left argument of the returned kernel k(X, Y)
        Y : array, shape (n_samples_Y, n_features), (optional, default=None)
            Right argument of the returned kernel k(X, Y). If None, k(X, X)
            if evaluated instead.
        eval_gradient : bool (optional, default=False)
            Determines whether the gradient with respect to the kernel
            hyperparameter is determined. Only supported when Y is None.
        active : np.ndarray (n_samples_X, n_features) (optional)
            Boolean array specifying which hyperparameters are active.
        Returns
        -------
        K : array, shape (n_samples_X, n_samples_Y)
            Kernel k(X, Y)
        K_gradient : array (opt.), shape (n_samples_X, n_samples_X, n_dims)
            The gradient of the kernel k(X, X) with respect to the
            hyperparameter of the kernel. Only returned when eval_gradient
            is True.
        """

        X = np.atleast_2d(X)

        if Y is not None and eval_gradient:
            raise ValueError("Gradient can only be evaluated when Y is None.")

        if Y is None:
            K = self.noise_level * np.eye(X.shape[0])

            if active is not None:
                K = K * active

            if eval_gradient:
                if not self.hyperparameter_noise_level.fixed:
                    return (K, self.noise_level * np.eye(X.shape[0])[:, :, np.newaxis])
                else:
                    return K, np.empty((X.shape[0], X.shape[0], 0))
            else:
                return K
        else:
            return np.zeros((X.shape[0], Y.shape[0]))


class HammingKernel(MagicMixin, kernels.HammingKernel):

    def __init__(
            self,
            length_scale: Union[float, Tuple[float, ...]] = 1.0,
            length_scale_bounds: Union[Tuple[float, float], List[Tuple[float, float]]] = (1e-5, 1e5),
            operate_on: Optional[np.ndarray] = None,
            prior: Optional[Prior] = None,
            has_conditions: bool = False,
    ) -> None:
        super(HammingKernel, self).__init__(length_scale=length_scale, length_scale_bounds=length_scale_bounds)
        self.set_active_dims(operate_on)
        self.prior = prior
        self.has_conditions = has_conditions

    def _call(
            self,
            X: np.ndarray,
            Y: Optional[np.ndarray] = None,
            eval_gradient: bool = False,
            active: Optional[np.ndarray] = None,
    ) -> Union[np.ndarray, Tuple[np.ndarray, np.ndarray]]:
        """Return the kernel k(X, Y) and optionally its gradient.
        Parameters
        ----------
        X : [array-like, shape=(n_samples_X, n_features)]
            Left argument of the returned kernel k(X, Y)
        Y : [array-like, shape=(n_samples_Y, n_features) or None(default)]
            Right argument of the returned kernel k(X, Y). If None, k(X, X)
            if evaluated instead.
        eval_gradient : [bool, False(default)]
            Determines whether the gradient with respect to the kernel
            hyperparameter is determined. Only supported when Y is None.
        active : np.ndarray (n_samples_X, n_features) (optional)
            Boolean array specifying which hyperparameters are active.
        Returns
        -------
        K : [array-like, shape=(n_samples_X, n_samples_Y)]
            Kernel k(X, Y)
        K_gradient : [array-like, shape=(n_samples_X, n_samples_X, n_dims)]
            The gradient of the kernel k(X, X) with respect to the
            hyperparameter of the kernel. Only returned when eval_gradient
            is True.
        Note
        ----
        Code partially copied from skopt (https://github.com/scikit-optimize).
        Made small changes to only compute necessary values and use scikit-learn helper functions.
        """

        X = np.atleast_2d(X)
        length_scale = sklearn.gaussian_process.kernels._check_length_scale(X, self.length_scale)

        if Y is None:
            Y = X
        elif eval_gradient:
            raise ValueError("gradient can be evaluated only when Y != X")
        else:
            Y = np.atleast_2d(Y)

        indicator = np.expand_dims(X, axis=1) != Y
        K = (-1 / (2 * length_scale ** 2) * indicator).sum(axis=2)
        K = np.exp(K)

        if active is not None:
            K = K * active

        if eval_gradient:
            # dK / d theta = (dK / dl) * (dl / d theta)
            # theta = log(l) => dl / d (theta) = e^theta = l
            # dK / d theta = l * dK / dl

            # dK / dL computation
            if np.iterable(length_scale) and length_scale.shape[0] > 1:
                grad = (np.expand_dims(K, axis=-1) * np.array(indicator, dtype=np.float32))
            else:
                grad = np.expand_dims(K * np.sum(indicator, axis=2), axis=-1)

            grad *= (1 / length_scale ** 3)

            return K, grad
        return K


from typing import DefaultDict
from .default_configuration_design import DefaultConfiguration
from .factorial_design import FactorialInitialDesign
from .latin_hypercube_design import LHDesign
from .random_design import RandomDesign
from .sobol_design import SobolDesign

ALL_avaliable_design = {
    'default': DefaultConfiguration,
    'fac': FactorialInitialDesign,
    'lh':LHDesign,
    'random': RandomDesign,
    'sobol': SobolDesign
}


from abc import ABC
import numpy as np
import warnings
from scipy import stats



class AbstractFeatureSpace(ABC):
    '''
    for optimizer using

    sparse array => features
    '''


    def __init__(self):
        pass

    # @abstractmethod
    def array_to_feature(self, array):
        '''
        array: expressed by configspace
        feature: expressed by optimizer
        '''
        pass

    # @abstractmethod
    def feature_to_array(self, feature):
        pass

class Identity():
    '''
    uniform to Gaussian
    '''

    def __init__(self):
        '''
        U(0, 1) to std gaussian feature
        define target feature's mean and std
        '''
        super().__init__()

    def sparse_array_to_feature(self, sparse_array):

        return sparse_array

    def feature_to_sparse_array(self, feature):
        return feature

class Cat2Onehot():

    def __init__(self):
        super().__init__()

    def sparse_array_to_feature(self, sparse_array, cat_num):
        '''
        sparse_array: int, index of max value
        return: one-hot code
        '''
        feat = np.zeros(cat_num)
        feat[int(sparse_array)] = 1
        return feat

    def feature_to_sparse_array(self, feature, cat_num):
        return np.argmax(feature).item()

class Ord2Uniform():
    '''
    ordinal to uniform(0, 1)
    '''

    def __init__(self):
        super().__init__()

    def sparse_array_to_feature(self, sparse_array, seqnum):
        '''
        sparse_array: int , one of 0, 1, 2 ... (seqnum-1)
        return a float in [0, 1]
        '''

        return sparse_array / (seqnum-1)

    def feature_to_sparse_array(self, feature, seqnum):
        return np.rint(feature * (seqnum-1))

class U2gaussian():
    '''
    uniform to Gaussian
    '''

    def __init__(self):
        '''
        U(0, 1) to std gaussian feature
        define target feature's mean and std
        '''
        super().__init__()

    def sparse_array_to_feature(self, sparse_array):
        '''
        convert to Gaussian distribution
        '''
        return stats.norm.ppf(sparse_array)

    def feature_to_sparse_array(self, feature):
        return stats.norm.cdf(feature)

class U2Onehot():
    def __init__(self):
        AbstractFeatureSpace.__init__(self)

    def sparse_array_to_feature(self, sparse_array, cat_num):
        '''
        sparse_array: 0~1
        return: one-hot code
        '''
        feat = np.zeros(cat_num)
        feat[np.uintp(sparse_array*cat_num)] = 1
        return feat

    def feature_to_sparse_array(self, feature, cat_num):
        return np.argmax(feature, -1).item() / (cat_num-1)


class Ordinal():

    def __init__(self):
        super().__init__()

    
    def feature_to_sparse_array(self, feature, seq_num):
        '''
        return int, one of 0, 1, 2 ... (seqnum-1)
        '''
        return threshold_discretization(feature, arity=seq_num)

    def sparse_array_to_feature(self, sparse_array, seq_num):
        '''
        feature： int , one of 0, 1, 2 ... (seqnum-1)
        return： N(0, 1)
        '''
        return inverse_threshold_discretization(sparse_array, arity=seq_num)


class Gaussian():

    def __init__(self, mean=0, std=1):
        '''
        std gaussian to specify gaussian
        define target feature's mean and std
        '''
        super().__init__()
        self.mean = mean
        self.std = std

    def sparse_array_to_feature(self, sparse_array):
        '''
        convert to Gaussian distribution
        '''
        return self.std * sparse_array + self.mean

    def feature_to_sparse_array(self, feature):
        return (feature - self.mean) / self.std

class Category():

    def __init__(self, deterministic=False):
        super().__init__()
        # self.cat_num = cat_num
        self.deterministic = deterministic

    def feature_to_sparse_array(self, feature, cat_num):
        '''
        return a int index
        '''

        return int(softmax_discretization(feature, cat_num, deterministic=self.deterministic))

    def sparse_array_to_feature(self, sparse_array, cat_num):
        '''
        convert to Gaussian distribution
        '''
        return inverse_softmax_discretization(sparse_array, cat_num)



def threshold_discretization(x, arity: int = 2):
    """Discretize by casting values from 0 to arity -1, assuming that x values
    follow a normal distribution.

    Parameters
    ----------
    x: list/array
       values to discretize
    arity: int
       the number of possible integer values (arity n will lead to values from 0 to n - 1)

    Note
    ----
    - nans are processed as negative infs (yields 0)
    """
    x = np.array(x, copy=True)
    if np.any(np.isnan(x)):
        warnings.warn("Encountered NaN values for discretization")
        x[np.isnan(x)] = -np.inf
    if arity == 2:  # special case, to have 0 yield 0
        return (np.array(x) > 0).astype(int)#.tolist()  # type: ignore
    else:
        return np.clip(arity * stats.norm.cdf(x), 0, arity - 1).astype(int)#.tolist()  # type: ignore


def inverse_threshold_discretization(indexes, arity: int = 2):
    '''
    to N(0,1)
    '''
    indexes_arr = np.array(indexes, copy=True)
    pdf_bin_size = 1 / arity
    # We take the center of each bin (in the pdf space)
    return stats.norm.ppf(indexes_arr * pdf_bin_size + (pdf_bin_size / 2))


def softmax_discretization(x, arity: int = 2, deterministic: bool = False):
    """Discretize a list of floats to a list of ints based on softmax probabilities.
    For arity n, a softmax is applied to the first n values, and the result
    serves as probability for the first output integer. The same process it
    applied to the other input values.

    Parameters
    ----------
    x: list/array
        the float values from a continuous space which need to be discretized
    arity: int
        the number of possible integer values (arity 2 will lead to values in {0, 1})
    deterministic: bool
        removes all randomness by returning the last mast probable output (highest values)

    Notes
    -----
    - if one or several inf values are present, only those are considered
    - in case of tie, the deterministic value is the first one (lowest) of the tie
    - nans and -infs are ignored, except if all are (then uniform random choice)
    """
    # data = np.array(x, copy=True, dtype=float).reshape((-1, arity))
    # if np.any(np.isnan(data)):
    #     warnings.warn("Encountered NaN values for discretization")
    #     data[np.isnan(data)] = -np.inf
    # if deterministic:
    #     output = np.argmax(data, axis=1).tolist()
    #     return output
    # return [np.random.choice(arity, p=softmax_probas(d)) for d in data]
    data = np.array(x, copy=True, dtype=float)
    if np.any(np.isnan(data)):
        warnings.warn("Encountered NaN values for discretization")
        data[np.isnan(data)] = -np.inf
    if deterministic:
        output = np.argmax(data, axis=1)
        return output
    return np.random.choice(arity, p=softmax_probas(data))
    # data = np.array(x, copy=True, dtype=float)#.reshape((-1, arity))
    # output = np.zeros_like(data)
    # if np.any(np.isnan(data)):
    #     warnings.warn("Encountered NaN values for discretization")
    #     data[np.isnan(data)] = -np.inf
    # if deterministic:
    #     max_idx = np.argmax(data, axis=1)
    #     output[np.arange(output.shape[0]), max_idx] = 1
    #     return output
    # for i, d in enumerate(data):
    #     max_idx = np.random.choice(arity, p=softmax_probas(d))
    #     output[i, max_idx] = 1
    # return output

def softmax_probas(data: np.ndarray) -> np.ndarray:
    # TODO: test directly? (currently through softmax discretization)
    # TODO: move nan case here?
    maxv = np.max(data)
    if np.abs(maxv) == np.inf or np.isnan(maxv):
        maxv = 0
    data = np.exp(data - maxv)
    if any(x == np.inf for x in data):  # deal with infinite positives special case
        data = np.array([int(x == np.inf) for x in data])
    if not sum(data):
        data = np.ones(len(data))
    return data / np.sum(data)

def inverse_softmax_discretization(_x: int, arity: int):
    # p is an arbitrary probability that the provided arg will be sampled with the returned point
    p = (1 / arity) * 1.5
    # x = np.zeros(arity)
    x = np.zeros(arity)
    x[int(_x)] = np.log((p * (arity - 1)) / (1 - p))
    return x


import numpy as np
from scipy import stats
from xbbo.acquisition_function.base import AbstractAcquisitionFunction

class EI():
    def __init__(self, surrogate, y_best):
        self.eta = 0.0
        self.surrogate = surrogate
        self.y_best = y_best

    def __call__(self, candidate): #
        mu, sigma = self.surrogate.predict_with_sigma(candidate)
        z = (self.y_best - mu - self.eta) / sigma
        ei = (self.y_best - mu -
              self.eta) * stats.norm.cdf(z) + sigma * stats.norm.pdf(z)
        return ei

    # def argmax(self, y_best, surrogate, candidates):
    #     best_ei = -1
    #     best_candidate = []
    #     for candidate in candidates:
    #         y_hat = surrogate.predict(candidate)
    #         ei = self._getEI(y_hat[0], y_hat[1], y_best)
    #         if ei > best_ei:
    #             best_ei = ei
    #             best_candidate = [candidate]
    #         elif ei == best_ei:
    #             best_candidate.append(candidate)
    #     return np.random.choice(best_candidate)

    def argmax(self, candidates):
        best_ei = -1
        # best_candidate = []
        candidates_rm_id = []
        # y_hats = list(zip(*surrogate.predict_with_sigma(candidates)))
        for i, candidate in enumerate(candidates):

            ei = self.__call__(candidate)
            if ei > best_ei:
                best_ei = ei
                # best_candidate = [candidate]
                candidates_rm_id = [i]
            elif ei == best_ei:
                # best_candidate.append(candidate)
                candidates_rm_id.append(i)

        assert candidates_rm_id
        idx = np.random.choice(len(candidates_rm_id))
        return candidates_rm_id[idx]

class EI_AcqFunc(AbstractAcquisitionFunction):
    def __init__(self, surrogate_model, rng):
        self.par = 0.0
        self.rng = rng
        super().__init__(surrogate_model)

    def argmax(self, candidates):
        print('depreciate!')
        # best_ei = -1
        # # best_candidate = []
        # candidates_rm_id = []
        # y_hats = list(zip(*surrogate.predict_with_sigma(candidates)))
        scores = self.__call__(candidates)
        
        return candidates[self.rng.choice(np.where(scores==scores.max())[0])]

    def _compute(self, X: np.ndarray, **kwargs) -> np.ndarray:
        """Computes the EI value and its derivatives.

        Parameters
        ----------
        X: np.ndarray(N, D), The input points where the acquisition function
            should be evaluated. The dimensionality of X is (N, D), with N as
            the number of points to evaluate at and D is the number of
            dimensions of one X.

        Returns
        -------
        np.ndarray(N,1)
            Expected Improvement of X
        """
        if len(X.shape) == 1:
            X = X[:, np.newaxis]

        m, v = self.surrogate_model.predict_marginalized_over_instances(X)
        s = np.sqrt(v)

        if self.y_best is None:
            raise ValueError('No current best specified. Call update('
                             'eta=<int>) to inform the acquisition function '
                             'about the current best value.')

        def calculate_f():
            z = (self.y_best - m - self.par) / s
            return (self.y_best - m - self.par) * stats.norm.cdf(z) + s * stats.norm.pdf(z)

        if np.any(s == 0.0):
            # if std is zero, we have observed x on all instances
            # using a RF, std should be never exactly 0.0
            # Avoid zero division by setting all zeros in s to one.
            # Consider the corresponding results in f to be zero.
            s_copy = np.copy(s)
            s[s_copy == 0.0] = 1.0
            f = calculate_f()
            f[s_copy == 0.0] = 0.0
        else:
            f = calculate_f()
        if (f < 0).any():
            raise ValueError(
                "Expected Improvement is smaller than 0 for at least one "
                "sample.")

        return f

# class MC_AcqFunc(AbstractAcquisitionFunction):
#     def __init__(self, surrogate_model, rng):
#         self.sample_num = 10
#         self.rng = rng
#         super().__init__(surrogate_model)

#     def _compute(self, X: np.ndarray) -> np.ndarray:
#         """Computes the EI value and its derivatives.

#         Parameters
#         ----------
#         X: np.ndarray(N, D), The input points where the acquisition function
#             should be evaluated. The dimensionality of X is (N, D), with N as
#             the number of points to evaluate at and D is the number of
#             dimensions of one X.

#         Returns
#         -------
#         np.ndarray(N,1)
#             Expected Improvement of X
#         """
#         if len(X.shape) == 1:
#             X = X[:, np.newaxis]
#         m, v = self.surrogate_model.predict_marginalized_over_instances(X, 'full_cov')
#         f = self.rng.multivariate_normal(m, v, size=self.sample_num)

#         return f # shape: (sample_num, N)



import os

import numpy as np
import pandas as pd
import matplotlib.pyplot as plt


class Record:

    def __init__(self, exp_dir):
        self.exp_dir = exp_dir
        # self.n_calls = cfg.OPTM.max_call
        self.features = [] # n_call * n_suggest * dim # 单独存储
        self.func_evals = [] # n_call * n_suggest * dim
        # self.losses = [] # n_call * n_suggest * 2
        # self.cum_min = {
        #     'cum_min_suggest_dict': [],
        #     'cum_min_losses': [],
        #     'cum_min_func_evals': []
        # }
        self.current_best = np.inf
        # self.metrics = [] # n_call * n_suggest * dim
        # self.budgets = []
        self.timing = { # 用来评估优化器的性能，如果是把model推理时间作为评估，应该放在loss、func_evals中
            'suggest_time_per_suggest': [],
            'observe_time_per_suggest': [],
            'eval_time_per_suggest': []
        }
        self.suggest_dict = []
        # index = pd.MultiIndex.from_product([self.n_calls * n_suggest, dim])
        self.df = None

    # def __str__(self):
    #     return np.asarray(self.func_evals).__str__()

    def size(self):
        return len(self.func_evals)

    def append(self, feature, y, suggest_point, timing):
        self.features.append(feature)
        self.func_evals.append(y)
        # self.losses.append(loss)
        for k in self.timing:
            self.timing[k].append(timing[k])
        self.suggest_dict.append(suggest_point)

    # def get_best(self):
    #     idx = np.argmin(np.asarray(self.func_evals)[...,0].ravel())
    #     return idx//len(self.func_evals[0]), \
    #            np.asarray(self.features).ravel()[idx],\
    #            np.asarray(self.func_evals)[...,0].ravel()[idx],\
    #            np.asarray(self.suggest_dict).ravel()[idx]

    # def plot(self):
    #     plt.plot(self.history)

    def is_duplicate(self, x, rtol=1e-5, atol=1e-8):
        return any(np.allclose(x_prev, x, rtol=rtol, atol=atol)
                   for x_prev in self.features)

    def save_to_file(self, r):
        if not os.path.exists(self.exp_dir):
            assert False
            # os.mkdir(cfg.GENARAL.exp_dir)
        if not os.path.exists(self.exp_dir + '/res'):
            os.mkdir(self.exp_dir + '/res/')
        self.func_evals = np.asarray(self.func_evals)
        # self.losses = np.asarray(self.losses)
        self.suggest_dict = np.asarray(self.suggest_dict)
        # reform = {
        #     (call, suggest): self.losses[call, suggest]
        #     for call in range(self.losses.shape[0])
        #     for suggest in range(self.losses.shape[1])
        # }
        self.df = pd.DataFrame(self.func_evals)#.T
        # self.losses_df.columns = pd.MultiIndex.from_product([['loss'], self.cfg.TEST_PROBLEM.losses])
        self.df.columns = ['func_evals']
        self.df['cum_func_evals'] = self.df['func_evals'].cummin()
        # self.losses_df.columns = cfg.TEST_PROBLEM.losses
        # self.losses_df.index.set_names(['call', 'suggest'], inplace=True)

        # reform = {
        #     (call, suggest): self.func_evals[call, suggest, :]
        #     for call in range(self.func_evals.shape[0])
        #     for suggest in range(self.func_evals.shape[1])
        # }
        # self.func_evals_df = pd.DataFrame(reform).T
        # self.func_evals_df.columns = pd.MultiIndex.from_product([['func_evals'], self.cfg.TEST_PROBLEM.func_evals])
        # # self.func_evals_df.columns = cfg.TEST_PROBLEM.func_evals
        # self.func_evals_df.index.set_names(['call', 'suggest'], inplace=True)

        # sd = pd.DataFrame(self.suggest_dict).stack()
        # sd.index.set_names(['call', 'suggest'], inplace=True)
        # sd.name = 'suggest_dict'
        # self.df = pd.concat([self.losses_df, self.func_evals_df], axis=1)
        # self.df['suggest_dict'] = sd

        self.time_df = pd.DataFrame(self.timing)
        self.time_df.index.set_names(['call'], inplace=True)

        # self.cum_min_df = pd.DataFrame(self.cum_min)
        # self.cum_min_df.index.set_names(['call'], inplace=True)
        self.df.to_csv(self.exp_dir + f'/res/res_{r}.csv')
        self.time_df.to_csv(self.exp_dir + f'/res/time_{r}.csv')
        # self.cum_min_df.to_csv(self.exp_dir + f'/res/cum_min_{r}.csv')

        self.features = np.asarray(self.features)
        np.savez(self.exp_dir + f'/res/array_{r}.npz', features=self.features)

        # if cfg.TEST_PROBLEM.metrics:
        #     reform = {
        #         (call, suggest): self.metrics[call, suggest, :]
        #         for call in range(self.metrics.shape[0])
        #         for suggest in range(self.metrics.shape[1])
        #     }
        #     self.metrics = pd.DataFrame(reform).T
        #     self.metrics.columns = cfg.TEST_PROBLEM.metrics
        # index = pd.MultiIndex.from_product([
        #     [f'call_{i}' for i in range(self.losses.shape[0])],
        #     [f'suggest_{i}' for i in range(self.losses.shape[1])]
        # ]
        # )


from typing import Optional
import numpy as np
from scipy import stats

from xbbo.acquisition_function.base import AbstractAcquisitionFunction


class MoGP_AcqFunc(AbstractAcquisitionFunction):
    def __init__(self, surrogate_model, base_models, rng, rho=None):
        self.par = 0.0
        self.base_models = base_models
        self.rng = rng
        self.rho = rho
        super().__init__(surrogate_model)

    def argmax(self, candidates):
        print('depreciate!')
        # best_ei = -1
        # # best_candidate = []
        # candidates_rm_id = []
        # y_hats = list(zip(*surrogate.predict_with_sigma(candidates)))
        scores = self.__call__(candidates)

        return candidates[self.rng.choice(np.where(scores == scores.max())[0])]

    def update_weight(self, w, rho=None):
        assert w.min() >= 0
        if rho:
            self.w = np.array(w, dtype='float')
            self.selfWeight = rho
        else:
            self.selfWeight = w[-1]
            self.w = np.array(w[:-1], dtype='float')
        #non_zero_weight_indices = (self.w**2 > 0).nonzero()[0]
        non_zero_weight_indices = self.w.nonzero()[0]
        self.pre_weight_model = []
        self.base_incuments = []
        for idx in non_zero_weight_indices:
            self.pre_weight_model.append(self.base_models[idx])
            self.base_incuments.append(self._base_incuments[idx])
        self.weight = self.w[non_zero_weight_indices]
        norm_coeff = (self.weight.sum() + self.selfWeight)
        self.weight /= norm_coeff
        self.selfWeight /= norm_coeff

    def _compute(self, X: np.ndarray) -> np.ndarray:
        """Computes the EI value and its derivatives.

        Parameters
        ----------
        X: np.ndarray(N, D), The input points where the acquisition function
            should be evaluated. The dimensionality of X is (N, D), with N as
            the number of points to evaluate at and D is the number of
            dimensions of one X.

        Returns
        -------
        np.ndarray(N,1)
            Expected Improvement of X
        """
        if len(X.shape) == 1:
            X = X[:, np.newaetas]

        m, v = self.surrogate_model.predict_marginalized_over_instances(X)
        s = np.sqrt(v)

        if self.y_best is None:
            raise ValueError('No current best specified. Call update('
                             'eta=<int>) to inform the acquisition function '
                             'about the current best value.')

        def calculate_f():
            z = (self.y_best - m - self.par) / s
            return (self.y_best - m -
                    self.par) * stats.norm.cdf(z) + s * stats.norm.pdf(z)

        def calculate_f_(y_best, m, s):
            z = (y_best - m - self.par) / s
            return (y_best - m -
                    self.par) * stats.norm.cdf(z) + s * stats.norm.pdf(z)

        if np.any(s == 0.0):
            # if std is zero, we have observed x on all instances
            # using a RF, std should be never exactly 0.0
            # Avoid zero division by setting all zeros in s to one.
            # Consider the corresponding results in f to be zero.
            s_copy = np.copy(s)
            s[s_copy == 0.0] = 1.0
            f = calculate_f()
            f[s_copy == 0.0] = 0.0
        else:
            f = calculate_f()
        if (f < 0).any():
            raise ValueError(
                "Expected Improvement is smaller than 0 for at least one "
                "sample.")
        # m, v = self.surrogate_model.predict_marginalized_over_instances(X)
        # denominator = self.selfWeight
        f *= self.selfWeight
        for d in range(len(self.weight)):
            m, v = self.base_models[d].predict_marginalized_over_instances(X)
            s = np.sqrt(v)
            y_best = self.base_incuments[d]

            if np.any(s == 0.0):
                # if std is zero, we have observed x on all instances
                # using a RF, std should be never exactly 0.0
                # Avoid zero division by setting all zeros in s to one.
                # Consider the corresponding results in f to be zero.
                s_copy = np.copy(s)
                s[s_copy == 0.0] = 1.0
                f_ = calculate_f_(y_best, m, v)
                f_[s_copy == 0.0] = 0.0
            else:
                f_ = calculate_f_(y_best, m, v)
            f += self.weight[d] * f_

        return f




'''
Reference: https://github.com/automl/DEHB
'''

from typing import List
import numpy as np

# from xbbo.configspace.feature_space import Uniform2Gaussian
from xbbo.search_algorithm.multi_fidelity.RFHB import RFHB
from xbbo.configspace.space import DenseConfiguration, DenseConfigurationSpace
from xbbo.core.trials import Trials, Trial
from xbbo.search_algorithm.multi_fidelity.utils.bracket_manager import BasicConfigGenerator, RFDEHB_ConfigGenerator, SHBracketManager
# from xbbo.search_algorithm.de_optimizer import DE
from xbbo.utils.constants import Key
from .. import alg_register

# class DEHB_CG(BasicConfigGenerator, DE):
#     def __init__(self, cs, budget, max_pop_size, rng, **kwargs) -> None:
#         BasicConfigGenerator.__init__(self, cs, budget, max_pop_size, rng, **kwargs)
#         DE.__init__(self, space=cs, init_budget=0,**kwargs)
#         self.reset(max_pop_size)

alg_marker = 'rfdehb'


@alg_register.register(alg_marker)
class RFDEHB(RFHB):
    name = alg_marker

    def __init__(
            self,
            space: DenseConfigurationSpace,
            budget_bound=[9, 729],
            #  mutation_factor=0.5,
            #  crossover_prob=0.5,
            #  strategy='rand1_bin',
            eta: int = 3,
            seed: int = 42,
            round_limit: int = 1,
            bracket_limit=np.inf,
            boundary_fix_type='random',
            encoding_cat='bin',
            encoding_ord='bin',
            **kwargs):
        RFHB.__init__(self,
                    space,
                    budget_bound,
                    eta,
                    seed=seed,
                    round_limit=round_limit,
                    bracket_limit=bracket_limit,
                    boundary_fix_type=boundary_fix_type,
                    encoding_cat=encoding_cat,
                    encoding_ord=encoding_ord,
                    **kwargs)

    def _observe(self, trial_list):
        for trial in trial_list:
            self.trials.add_a_trial(trial, permit_duplicate=True)
            fitness = trial.observe_value
            job_info = trial.info
            budget = job_info[Key.BUDGET]
            parent_id = job_info['parent_id']
            individual = trial.array  # TODO
            for bracket in self.active_brackets:
                if bracket.bracket_id == job_info['bracket_id']:
                    # registering is IMPORTANT for Bracket Manager to perform SH
                    bracket.register_job(budget)  # may be new row
                    # bracket job complete
                    bracket.complete_job(
                        budget)  # IMPORTANT to perform synchronous SH
            # carry out DE selection
            if fitness <= self.cg[budget].population_fitness[parent_id]:  # TODO
                self.cg[budget].population[parent_id] = individual
                self.cg[budget].population_fitness[parent_id] = fitness
                # updating incumbents
                if fitness < self.current_best_fitness:
                    self.current_best = individual
                    self.current_best_fitness = trial.observe_value
                    self.current_best_trial = trial

        self._clean_inactive_brackets()
        # if len(self.active_brackets) == 0:  # complete current bracket
        #     self.round_recoder = self.bracket_counter // self.max_SH_iter

    def _init_subpop(self, **kwargs):
        """ List of DE objects corresponding to the budgets (fidelities)
        """
        self.cg = {}
        for i, b in enumerate(self._max_pop_size.keys()):
            self.cg[b] = RFDEHB_ConfigGenerator(
                self.space,
                budget=b,
                max_pop_size=self._max_pop_size[b],
                rng=self.rng,
                eta = self.eta,
                **kwargs)

    def _acquire_candidate(self, bracket, budget):
        """ Generates/chooses a configuration based on the budget and iteration number
        """
        # select a parent/target
        parent_id = self._get_next_idx_for_subpop(budget, bracket)
        target = self.cg[budget].population[parent_id]
        lower_budget, num_configs = bracket.get_lower_budget_promotions(budget)
        # if bracket.is_new_rung():
        #     self.cg[budget].population_fitness[num_configs:] = np.inf
        # Fix bugs in the original author's implementation code
        # if self.bracket_counter == 0 and budget != bracket.budgets[0]:
        if self.bracket_counter < self.max_SH_iter and budget != bracket.budgets[
                0]:
            # 第一列，第二行开始，直接挑选最优的进入下一轮
            # TODO: check if generalizes to all budget spacings
            individual = self._get_promotion_candidate(lower_budget, budget,
                                                       num_configs)
            individual = self.fix_boundary(individual)
            return individual, parent_id
            # else: # 每一列中的第一行，随机生成config
        mutation_pop_idx = np.argsort(
            self.cg[lower_budget].population_fitness)[:num_configs]
        # mutation_pop_idx = np.argpartition(self.cg[lower_budget].population_fitness, num_configs-1)[:num_configs]
        mutation_pop = self.cg[lower_budget].population[mutation_pop_idx]
        # generate mutants from previous budget subpopulation or global population
        if len(mutation_pop) < self.cg[budget]._min_pop_size:
            filler = self.cg[budget]._min_pop_size - len(mutation_pop) + 1
            new_pop = self.cg[budget]._init_mutant_population(
                pop_size=filler,
                population=self._concat_pops(),
                target=target,
                best=self.current_best)
            mutation_pop = np.concatenate((mutation_pop, new_pop))
        individual = self.cg[budget].batch_select_mut_cross_child(current=target, best=self.current_best, alt_pop=mutation_pop, num_configs=num_configs)


        individual = self.fix_boundary(individual)
        return individual, parent_id

    def _get_promotion_candidate(self, low_budget, high_budget, n_configs):
        """ Manages the population to be promoted from the lower to the higher budget.

        This is triggered or in action only during the first full HB bracket, which is equivalent
        to the number of brackets <= max_SH_iter.
        """
        # finding the individuals that have been evaluated (fitness < np.inf)
        evaluated_configs = np.where(
            self.cg[low_budget].population_fitness != np.inf)[0]
        promotion_candidate_pop = self.cg[low_budget].population[
            evaluated_configs]
        promotion_candidate_fitness = self.cg[low_budget].population_fitness[
            evaluated_configs]
        # ordering the evaluated individuals based on their fitness values
        pop_idx = np.argsort(promotion_candidate_fitness)
        labels = promotion_candidate_fitness.copy() 
        # creating population for promotion if none promoted yet or nothing to promote
        if self.cg[high_budget].promotion_pop is None or \
                len(self.cg[high_budget].promotion_pop) == 0:
            self.cg[high_budget].promotion_pop = np.empty((0, self.dimension))
            self.cg[high_budget].promotion_fitness = np.array([])

            # iterating over the evaluated individuals from the lower budget and including them
            # in the promotion population for the higher budget only if it's not in the population
            # this is done to ensure diversity of population and avoid redundant evaluations
            for idx in pop_idx:
                individual = promotion_candidate_pop[idx]
                # checks if the candidate individual already exists in the high budget population
                if np.any(
                        np.all(individual == self.cg[high_budget].population,
                               axis=1)):
                    # skipping already present individual to allow diversity and reduce redundancy
                    continue
                self.cg[high_budget].promotion_pop = np.append(
                    self.cg[high_budget].promotion_pop, [individual], axis=0)
                self.cg[high_budget].promotion_fitness = np.append(
                    self.cg[high_budget].promotion_pop,
                    promotion_candidate_fitness[pop_idx])
            # retaining only n_configs
            self.cg[high_budget].promotion_pop = self.cg[
                high_budget].promotion_pop[:n_configs]
            self.cg[high_budget].promotion_fitness = \
                self.cg[high_budget].promotion_fitness[:n_configs]

        if len(self.cg[high_budget].promotion_pop) > 0:
            config = self.cg[high_budget].promotion_pop[0]
            # removing selected configuration from population
            self.cg[high_budget].promotion_pop = self.cg[
                high_budget].promotion_pop[1:]
            self.cg[high_budget].promotion_fitness = self.cg[
                high_budget].promotion_fitness[1:]
        else:
            # in case of an edge failure case where all high budget individuals are same
            # just choose the best performing individual from the lower budget (again)
            config = self.cg[low_budget].population[pop_idx[0]]
        self.cg[low_budget].add_fitting(promotion_candidate_pop, labels)
        return config

    def _get_next_idx_for_subpop(self, budget, bracket):
        """ Maintains a looping counter over a subpopulation, to iteratively select a parent
        """
        parent_id = self.cg[budget].parent_idx
        self.cg[budget].parent_idx += 1
        self.cg[budget].parent_idx = self.cg[
            budget].parent_idx % self._max_pop_size[budget]
        return parent_id

    def _concat_pops(self, exclude_budget=None):
        """ Concatenates all subpopulations
        """
        budgets = list(self.budgets)
        if exclude_budget is not None:
            budgets.remove(exclude_budget)
        pop = []
        for budget in budgets:
            pop.extend(self.cg[budget].population.tolist())
        return np.array(pop)


opt_class = RFDEHB



from typing import Dict, List, Optional, Tuple, Union
import typing
import numpy as np
from ConfigSpace.hyperparameters import \
    CategoricalHyperparameter, UniformFloatHyperparameter, UniformIntegerHyperparameter,Constant
import sklearn.gaussian_process.kernels
from sklearn.decomposition import PCA
from sklearn.preprocessing import MinMaxScaler
from sklearn.exceptions import NotFittedError
from skopt.learning.gaussian_process.kernels import Kernel
from skopt.learning.gaussian_process import GaussianProcessRegressor

from xbbo.configspace.space import DenseConfigurationSpace
from xbbo.surrogate.gp_prior import Prior, SoftTopHatPrior, TophatPrior


class SurrogateModel(object):
    """Abstract implementation of the Model API.

    **Note:** The input dimensionality of Y for training and the output dimensions
    of all predictions (also called ``n_objectives``) depends on the concrete
    implementation of this abstract class.

    Attributes
    ----------
    instance_features : np.ndarray(I, K)
        Contains the K dimensional instance features
        of the I different instances
    pca : sklearn.decomposition.PCA
        Object to perform PCA
    pca_components : float
        Number of components to keep or None
    n_feats : int
        Number of instance features
    n_params : int
        Number of parameters in a configuration (only available after train has
        been called)
    scaler : sklearn.preprocessing.MinMaxScaler
        Object to scale data to be withing [0, 1]
    var_threshold : float
        Lower bound vor variance. If estimated variance < var_threshold, the set
        to var_threshold
    types : list
        If set, contains a list with feature types (cat,const) of input vector
    """
    def __init__(self,
                 types: np.ndarray,
                 bounds: typing.List[typing.Tuple[float, float]],
                 instance_features: np.ndarray = None,
                 pca_components: float = None,
                 **kwargs):
        """Constructor

        Parameters
        ----------
        types : np.ndarray (D)
            Specifies the number of categorical values of an input dimension where
            the i-th entry corresponds to the i-th input dimension. Let's say we
            have 2 dimension where the first dimension consists of 3 different
            categorical choices and the second dimension is continuous than we
            have to pass np.array([2, 0]). Note that we count starting from 0.
        bounds : list
            Specifies the bounds for continuous features.
        instance_features : np.ndarray (I, K)
            Contains the K dimensional instance features
            of the I different instances
        pca_components : float
            Number of components to keep when using PCA to reduce
            dimensionality of instance features. Requires to
            set n_feats (> pca_dims).
        """
        self.instance_features = instance_features
        self.pca_components = pca_components

        if instance_features is not None:
            self.n_feats = instance_features.shape[1]
        else:
            self.n_feats = 0

        self.n_params = None  # will be updated on train()

        self.pca = None
        self.scaler = None
        if self.pca_components and self.n_feats > self.pca_components:
            self.pca = PCA(n_components=self.pca_components)
            self.scaler = MinMaxScaler()

        # Never use a lower variance than this
        self.var_threshold = 10**-5

        self.bounds = bounds
        self.types = types
        # Initial types array which is used to reset the type array at every call to train()
        self._initial_types = types.copy()
        self.do_optimize = kwargs.get('do_optimize', True)

    def train(self, X: np.ndarray, Y: np.ndarray) -> 'SurrogateModel':
        """Trains the Model on X and Y.

        Parameters
        ----------
        X : np.ndarray [n_samples, n_features (config + instance features)]
            Input data points.
        Y : np.ndarray [n_samples, n_objectives]
            The corresponding target values. n_objectives must match the
            number of target names specified in the constructor.

        Returns
        -------
        self : AbstractModel
        """
        self.types = self._initial_types.copy()

        if len(X.shape) != 2:
            raise ValueError('Expected 2d array, got %dd array!' %
                             len(X.shape))
        if X.shape[1] != len(self.types):
            raise ValueError(
                'Feature mismatch: X should have %d features, but has %d' %
                (X.shape[1], len(self.types)))
        if X.shape[0] != Y.shape[0]:
            raise ValueError('X.shape[0] (%s) != y.shape[0] (%s)' %
                             (X.shape[0], Y.shape[0]))

        self.n_params = X.shape[1] - self.n_feats

        # reduce dimensionality of features of larger than PCA_DIM
        if self.pca and X.shape[0] > self.pca.n_components:
            X_feats = X[:, -self.n_feats:]
            # scale features
            X_feats = self.scaler.fit_transform(X_feats)
            X_feats = np.nan_to_num(X_feats)  # if features with max == min
            # PCA
            X_feats = self.pca.fit_transform(X_feats)
            X = np.hstack((X[:, :self.n_params], X_feats))
            if hasattr(self, "types"):
                # for RF, adapt types list
                # if X_feats.shape[0] < self.pca, X_feats.shape[1] ==
                # X_feats.shape[0]
                self.types = np.array(
                    np.hstack((self.types[:self.n_params],
                               np.zeros((X_feats.shape[1])))),
                    dtype=np.uint,
                )

        return self._train(X, Y)

    def _train(self, X: np.ndarray, Y: np.ndarray, **kwargs) -> 'SurrogateModel':
        """Trains the random forest on X and y.

        Parameters
        ----------
        X : np.ndarray [n_samples, n_features (config + instance features)]
            Input data points.
        Y : np.ndarray [n_samples, n_objectives]
            The corresponding target values. n_objectives must match the
            number of target names specified in the constructor.

        Returns
        -------
        self
        """
        raise NotImplementedError

    def predict(
        self,
        X: np.ndarray,
        cov_return_type: typing.Optional[str] = 'diagonal_cov'
    ) -> typing.Tuple[np.ndarray, np.ndarray]:
        """
        Predict means and variances for given X.

        Parameters
        ----------
        X : np.ndarray of shape = [n_samples, n_features (config + instance features)]
            Training samples

        Returns
        -------
        means : np.ndarray of shape = [n_samples, n_objectives]
            Predictive mean
        vars : np.ndarray  of shape = [n_samples, n_objectives]
            Predictive variance
        """
        import warnings
        with warnings.catch_warnings():
            warnings.simplefilter("ignore", category=UserWarning)

            if len(X.shape) != 2:
                raise ValueError('Expected 2d array, got %dd array!' %
                                 len(X.shape))
            if X.shape[1] != len(self._initial_types):
                raise ValueError(
                    'Rows in X should have %d entries but have %d!' %
                    (len(self._initial_types), X.shape[1]))

            if self.pca:
                try:
                    X_feats = X[:, -self.n_feats:]
                    X_feats = self.scaler.transform(X_feats)
                    X_feats = self.pca.transform(X_feats)
                    X = np.hstack((X[:, :self.n_params], X_feats))
                except NotFittedError:
                    pass  # PCA not fitted if only one training sample

            if X.shape[1] != len(self.types):
                raise ValueError(
                    'Rows in X should have %d entries but have %d!' %
                    (len(self.types), X.shape[1]))

            mean, var = self._predict(X, cov_return_type=cov_return_type)
            if cov_return_type is None:
                return mean, var
            if len(mean.shape) == 1:
                mean = mean.reshape((-1, 1))
            if len(var.shape) == 1:
                var = var.reshape((-1, 1))

            return mean, var

    def _predict(
        self,
        X: np.ndarray,
        cov_return_type: typing.Optional[str] = 'diagonal_cov'
    ) -> typing.Tuple[np.ndarray, np.ndarray]:
        """
        Predict means and variances for given X.

        Parameters
        ----------
        X : np.ndarray
            [n_samples, n_features (config + instance features)]

        Returns
        -------
        means : np.ndarray of shape = [n_samples, n_objectives]
            Predictive mean
        vars : np.ndarray  of shape = [n_samples, n_objectives]
            Predictive variance
        """
        raise NotImplementedError()

    def predict_marginalized_over_instances(
            self, X: np.ndarray, cov_return_type: typing.Optional[str] = 'diagonal_cov') -> typing.Tuple[np.ndarray, np.ndarray]:
        """Predict mean and variance marginalized over all instances.

        Returns the predictive mean and variance marginalised over all
        instances for a set of configurations.

        Parameters
        ----------
        X : np.ndarray
            [n_samples, n_features (config)]

        Returns
        -------
        means : np.ndarray of shape = [n_samples, 1]
            Predictive mean
        vars : np.ndarray  of shape = [n_samples, 1]
            Predictive variance
        """

        if len(X.shape) != 2:
            raise ValueError('Expected 2d array, got %dd array!' %
                             len(X.shape))
        if X.shape[1] != len(self.types):
            raise ValueError('Rows in X should have %d entries but have %d!' %
                             (len(self.types), X.shape[1]))

        if self.instance_features is None or \
                len(self.instance_features) == 0:
            mean, var = self.predict(X, cov_return_type)
            var[var < self.var_threshold] = self.var_threshold
            var[np.isnan(var)] = self.var_threshold
            return mean, var
        else:
            n_instances = len(self.instance_features)

        mean = np.zeros(X.shape[0])
        var = np.zeros(X.shape[0])
        for i, x in enumerate(X):
            X_ = np.hstack((np.tile(x,
                                    (n_instances, 1)), self.instance_features))
            means, vars = self.predict(X_, cov_return_type)
            # VAR[1/n (X_1 + ... + X_n)] =
            # 1/n^2 * ( VAR(X_1) + ... + VAR(X_n))
            # for independent X_1 ... X_n
            var_x = np.sum(vars) / (len(vars)**2)
            if var_x < self.var_threshold:
                var_x = self.var_threshold

            var[i] = var_x
            mean[i] = np.mean(means)

        if len(mean.shape) == 1:
            mean = mean.reshape((-1, 1))
        if len(var.shape) == 1:
            var = var.reshape((-1, 1))

        return mean, var
    
    def update_weight(self, w, rho=None):
        pass


class BaseGP(SurrogateModel):
    def __init__(self,
                 configspace: DenseConfigurationSpace,
                 types: List[int],
                 bounds: List[Tuple[float, float]],
                 rng: np.random.RandomState,
                 normalize_y: bool = True,
                 instance_features: Optional[np.ndarray] = None,
                 pca_components: Optional[int] = None,
                 **kwargs):
        """
        Abstract base class for all Gaussian process models.
        """
        super().__init__(types=types,
                         bounds=bounds,
                         instance_features=instance_features,
                         pca_components=pca_components,
                         **kwargs)

        self.configspace = configspace
        self.rng = rng
        self.normalize_y = normalize_y
        kernel = kwargs.get('kernel')
        self.kernel = kernel if kernel else self._get_kernel()
        self.gp = self._get_gp()
        self.conditional = dict()  # type: Dict[int, bool]
        self.impute_values = dict()  # type: Dict[int, float]

    def _get_kernel(self) -> Kernel:
        raise NotImplementedError()

    def _get_gp(self) -> GaussianProcessRegressor:
        raise NotImplementedError()

    def _normalize_y(self, y: np.ndarray) -> np.ndarray:
        """Normalize data to zero mean unit standard deviation.

        Parameters
        ----------
        y : np.ndarray
            Targets for the Gaussian process

        Returns
        -------
        np.ndarray
        """
        self.mean_y_ = np.mean(y)
        self.std_y_ = np.std(y)
        if self.std_y_ == 0:
            self.std_y_ = 1
        return (y - self.mean_y_) / self.std_y_

    def _untransform_y(
        self,
        y: np.ndarray,
        var: Optional[np.ndarray] = None,
    ) -> Union[np.ndarray, Tuple[np.ndarray, np.ndarray]]:
        """Transform zeromean unit standard deviation data into the regular space.

        This function should be used after a prediction with the Gaussian process which was trained on normalized data.

        Parameters
        ----------
        y : np.ndarray
            Normalized data.
        var : np.ndarray (optional)
            Normalized variance

        Returns
        -------
        np.ndarray on Tuple[np.ndarray, np.ndarray]
        """
        y = y * self.std_y_ + self.mean_y_
        if var is not None:
            var = var * self.std_y_**2
            return y, var
        return y

    def _get_all_priors(
        self,
        add_bound_priors: bool = True,
        add_soft_bounds: bool = False,
    ) -> List[List[Prior]]:
        # Obtain a list of all priors for each tunable hyperparameter of the kernel
        all_priors = []
        to_visit = []
        # to_visit.append(self.gp.kernel.k1)
        # to_visit.append(self.gp.kernel.k2)
        to_visit.append(self.gp.kernel)  # fix single kernel
        while len(to_visit) > 0:
            current_param = to_visit.pop(0)
            if isinstance(current_param,
                          sklearn.gaussian_process.kernels.KernelOperator):
                to_visit.insert(0, current_param.k1)
                to_visit.insert(1, current_param.k2)
                continue
            elif isinstance(current_param,
                            sklearn.gaussian_process.kernels.Kernel):
                hps = current_param.hyperparameters
                assert len(hps) == 1
                hp = hps[0]
                if hp.fixed:
                    continue
                bounds = hps[0].bounds
                for i in range(hps[0].n_elements):
                    priors_for_hp = []
                    if current_param.prior is not None:
                        priors_for_hp.append(current_param.prior)
                    if add_bound_priors:
                        if add_soft_bounds:
                            priors_for_hp.append(
                                SoftTopHatPrior(
                                    lower_bound=bounds[i][0],
                                    upper_bound=bounds[i][1],
                                    rng=self.rng,
                                    exponent=2,
                                ))
                        else:
                            priors_for_hp.append(
                                TophatPrior(
                                    lower_bound=bounds[i][0],
                                    upper_bound=bounds[i][1],
                                    rng=self.rng,
                                ))
                    all_priors.append(priors_for_hp)
        return all_priors

    def _set_has_conditions(self) -> None:
        has_conditions = len(self.configspace.get_conditions()) > 0
        to_visit = []
        to_visit.append(self.kernel)
        while len(to_visit) > 0:
            current_param = to_visit.pop(0)
            if isinstance(current_param,
                          sklearn.gaussian_process.kernels.KernelOperator):
                to_visit.insert(0, current_param.k1)
                to_visit.insert(1, current_param.k2)
                current_param.has_conditions = has_conditions
            elif isinstance(current_param,
                            sklearn.gaussian_process.kernels.Kernel):
                current_param.has_conditions = has_conditions
            else:
                raise ValueError(current_param)

    # def _impute_inactive(self, X: np.ndarray) -> np.ndarray:
    #     X = X.copy()
    #     X[~np.isfinite(X)] = -1
    #     return X
    def _impute_inactive(self, X: np.ndarray) -> np.ndarray:
        X = X.copy()
        X[~np.isfinite(X)] = -1
        for idx, hp in enumerate(self.configspace.get_hyperparameters()):
            if idx not in self.conditional:
                parents = self.configspace.get_parents_of(hp.name)
                if len(parents) == 0:
                    self.conditional[idx] = False
                else:
                    self.conditional[idx] = True
                    if isinstance(hp, CategoricalHyperparameter):
                        self.impute_values[idx] = len(hp.choices)
                    elif isinstance(hp, (UniformFloatHyperparameter,
                                         UniformIntegerHyperparameter)):
                        self.impute_values[idx] = -1
                    elif isinstance(hp, Constant):
                        self.impute_values[idx] = 1
                    else:
                        raise ValueError

            if self.conditional[idx] is True:
                nonfinite_mask = ~np.isfinite(X[:, idx])
                X[nonfinite_mask, idx] = self.impute_values[idx]

        return X




class BaseRF(SurrogateModel):
    def __init__(self,
                 configspace: DenseConfigurationSpace,
                 types: typing.List[int],
                 bounds: List[Tuple[float, float]],
                 instance_features: Optional[np.ndarray] = None,
                 pca_components: Optional[int] = None,
                 **kwargs) -> None:
        """
        Abstract base class for all random forest models.
        """
        self.configspace = configspace
        super().__init__(types=types,
                         bounds=bounds,
                         instance_features=instance_features,
                         pca_components=pca_components,
                         **kwargs)

        self.conditional = dict()  # type: Dict[int, bool]
        self.impute_values = dict()  # type: Dict[int, float]

    def _impute_inactive(self, X: np.ndarray) -> np.ndarray:
        X = X.copy()
        for idx, hp in enumerate(self.configspace.get_hyperparameters()):
            if idx not in self.conditional:
                parents = self.configspace.get_parents_of(hp.name)
                if len(parents) == 0:
                    self.conditional[idx] = False
                else:
                    self.conditional[idx] = True
                    if isinstance(hp, CategoricalHyperparameter):
                        self.impute_values[idx] = len(hp.choices)
                    elif isinstance(hp, (UniformFloatHyperparameter,
                                         UniformIntegerHyperparameter)):
                        self.impute_values[idx] = -1
                    elif isinstance(hp, Constant):
                        self.impute_values[idx] = 1
                    else:
                        raise ValueError

            if self.conditional[idx] is True:
                nonfinite_mask = ~np.isfinite(X[:, idx])
                X[nonfinite_mask, idx] = self.impute_values[idx]

        return X



import numpy as np
import gc

from xbbo.pipeline.bbo_benchmark import BBObenchmark
from xbbo.utils.analysis import Analyse, Analyse_multi_benchmark
from xbbo.utils.config import cfg, load_cfg_fom_args
from xbbo.utils.constants import MAXINT
from xbbo.utils.util import dumpJson

exp = []
def do_experiment(cfg_clone):  # pragma: main
    # seed = cfg.GENARAL.random_seed
    SEED = cfg_clone.GENERAL.random_seed
    rng = np.random.RandomState(SEED)
    if cfg_clone.GENERAL.pipeline == 'BBO':
        # bbo = BBObenchmark(cfg_clone, SEED)

        # for r in range(cfg_clone.repeat_num):
        #     print('==EXP {}==:'.format(r))
        #     # np.random.seed(SEED)
        #     # random.seed(SEED)

        #     bbo.run_one_exp()
        #     bbo.save_to_file(r)
        #     print('=' * 20)
        #     seed = rng.randint(MAXINT)
        #     bbo.reset(seed)
        #     gc.collect()

        for r in range(cfg_clone.repeat_num):
            print('==EXP{}-{}-dir:"{}"==:'.format(r,cfg_clone.OPTM.name,cfg_clone.GENERAL.exp_dir))
            # np.random.seed(SEED)
            # random.seed(SEED)
            seed = rng.randint(MAXINT)
            bbo = BBObenchmark(cfg_clone, seed)

            bbo.run_one_exp()
            bbo.save_to_file(r)
            # exp.append(bbo.optimizer_instance.exp_selection_success)
            print('='*20)
        # dumpJson('./', 'exp.json', exp)
    else:
        raise NotImplementedError


if __name__ == '__main__':
    confs = {
        # "./cfgs/ext_openbox_bohb.yaml": ["--mark", "openbox_bohb"],
        # "./cfgs/ext_openbox_hb.yaml": ["--mark", "openbox_hb"],
        # "./cfgs/rfdehb.yaml": ["--mark", "rfdehb"],
        # "./cfgs/ext_dehb.yaml": ["--mark", "ext_dehb"],
        # "./cfgs/rfhb.yaml": ["--mark", "RFHB"],
        # "./cfgs/dehb.yaml": ["--mark", "DEHB"],
        # # "./cfgs/rs.yaml": ["--mark", "RS"],
        # "./cfgs/bohb.yaml": ["--mark", "bohb"],
        "./cfgs/mfes-bohb.yaml": ["--mark", "mfes-bohb"],
        # "./cfgs/ext_openbox_mfes.yaml": ["--mark", "openbox-mfes"],
        "./cfgs/hb.yaml": ["--mark", "hb"],
        # "./cfgs/ext_hb.yaml": ["--mark", "ext_hb"],
        # "./cfgs/ext_bohb.yaml": ["--mark", "ext_bohb"],
    }
    general_argv = ["-r", "5"]
    general_opts = ["TEST_PROBLEM.name", "countingones"]
    for conf in confs:
        cfg_clone = cfg.clone()
        cfg.freeze()

        argv = ["-c", conf]
        argv.extend(general_argv)
        argv.extend(confs[conf])
        argv.extend(general_opts)
        load_cfg_fom_args(cfg_clone,
                          argv=argv)  # repeat 3 times with diffent seeds
        cfg_clone.OPTM.kwargs.bracket_limit=20
        cfg_clone.OPTM.kwargs.round_limit=4
        do_experiment(cfg_clone)
        cfg.defrost()

    # cfg_clone = cfg.clone()
    # cfg.freeze()
    # load_cfg_fom_args(cfg_clone,
    #                   argv=[
    #                       "-c", "./cfgs/rfhb.yaml", "-r", "50", "--mark",
    #                       "RFHB", "TEST_PROBLEM.name", "nas_201"
    #                   ])  # repeat 3 times with diffent seeds
    # do_experiment(cfg_clone)
    # cfg.defrost()

    # cfg_clone = cfg.clone()
    # cfg.freeze()
    # load_cfg_fom_args(
    #     cfg_clone,
    #     argv=["-c", "./cfgs/dehb.yaml", "-r", "50", "--mark",
    #           "DEHB"])  # repeat 3 times with diffent seeds
    # do_experiment(cfg_clone)
    # cfg.defrost()

    # cfg_clone = cfg.clone()
    # cfg.freeze()
    # load_cfg_fom_args(
    #     cfg_clone, argv=["-c", "./cfgs/rs.yaml", "-r", "50", "--mark",
    #                      "RS"])  # repeat 3 times with diffent seeds
    # do_experiment(cfg_clone)
    # cfg.defrost()

    # # # cfg_clone = cfg.clone()
    # # # cfg.freeze()
    # # # load_cfg_fom_args(cfg_clone, argv=[                "-c",
    # # #             "./cfgs/smac3.yaml",
    # # #             "-r",
    # # #             "50",
    # # #             "--mark",
    # # #             "smac3"])  # repeat 3 times with diffent seeds

    # cfg_clone = cfg.clone()
    # cfg.freeze()
    # load_cfg_fom_args(cfg_clone,
    #                   argv=[
    #                       "-c", "./cfgs/bohb.yaml", "-r", "50", "--mark",
    #                       "bohb_array_inf"
    #                   ])  # repeat 3 times with diffent seeds
    # do_experiment(cfg_clone)
    # cfg.defrost()

    # cfg_clone = cfg.clone()
    # cfg.freeze()
    # load_cfg_fom_args(
    #     cfg_clone, argv=["-c", "./cfgs/hb.yaml", "-r", "50", "--mark",
    #                      "hb"])  # repeat 3 times with diffent seeds
    # do_experiment(cfg_clone)
    # cfg.defrost()
    marks = ["hb","DEHB","bohb", "ext_hb", "ext_bohb","ext_dehb", "openbox_hb","openbox_bohb", "mfes-bohb", "openbox-mfes"]
    # marks = ["DEHB_DEHB", "RFHB", "hb","DEHB","bohb_array_inf", "RFHB_OH"]
    Analyse('./exp', benchmark='countingones', marks=marks, legend_size=16)
    # Analyse_multi_benchmark()



# License: 3-clause BSD
# Copyright (c) 2016-2018, Ml4AAD Group (http://www.ml4aad.org/)

from typing import List

from xbbo.configspace.space import DenseConfiguration
from xbbo.initial_design.base import InitialDesign


class DefaultConfiguration(InitialDesign):

    """Initial design that evaluates default configuration"""

    def _select_configurations(self, num=None) -> List[DenseConfiguration]:
        """Selects the default configuration.

        Returns
        -------
        config: Configuration
            Initial incumbent configuration.
        """

        config = self.cs.get_default_configuration()
        config.origin = 'Default'
        self.init_budget = 1
        return [config]



from collections import OrderedDict
import typing
import numpy as np
from ConfigSpace.hyperparameters import NumericalHyperparameter, \
    Constant, CategoricalHyperparameter, OrdinalHyperparameter

from xbbo.configspace.space import DenseConfiguration, DenseConfigurationSpace, deactivate_inactive_hyperparameters


class InitialDesign:
    '''
    reference: https://github.com/automl/SMAC3/blob/master/xbbo/initial_design/initial_design.py
    '''
    def __init__(self,
                 cs: DenseConfigurationSpace,
                 rng: np.random.RandomState,
                 ta_run_limit: typing.Optional[int] = None,
                 n_configs_x_params: typing.Optional[int] = 10,
                 init_budget: typing.Optional[int] = None,
                 max_config_fracs: float = 0.25,
                 **kwargs) -> None:
        self.cs = cs
        self.dim = cs.get_dimensions()
        self.rng = rng
        # n_params = len(self.cs.get_hyperparameters())
        if init_budget is not None:
            self.init_budget = init_budget
        elif n_configs_x_params is not None:
            self.init_budget = int(
                max(
                    1,
                    min(n_configs_x_params * self.dim,
                        (max_config_fracs * ta_run_limit))))

    def select_configurations(self) -> typing.List[DenseConfiguration]:
        self.configs = self._select_configurations()
        self.configs = list(OrderedDict.fromkeys(self.configs))
        return self.configs

    def _select_configurations(self, num=None) -> typing.List[DenseConfiguration]:
        raise NotImplementedError

    def _transform_continuous_designs(
            self, design: np.ndarray, origin: str,
            cs: DenseConfigurationSpace) -> typing.List[DenseConfiguration]:

        # params = cs.get_hyperparameters()
        # for idx, param in enumerate(params):
        #     if isinstance(param, NumericalHyperparameter):
        #         continue
        #     elif isinstance(param, Constant):
        #         # add a vector with zeros
        #         design_ = np.zeros(np.array(design.shape) + np.array((0, 1)))
        #         design_[:, :idx] = design[:, :idx]
        #         design_[:, idx + 1:] = design[:, idx:]
        #         design = design_
        #     elif isinstance(param, CategoricalHyperparameter):
        #         v_design = design[:, idx]
        #         v_design[v_design == 1] = 1 - 10**-10
        #         design[:, idx] = np.array(v_design * len(param.choices),
        #                                   dtype=np.int)
        #     elif isinstance(param, OrdinalHyperparameter):
        #         v_design = design[:, idx]
        #         v_design[v_design == 1] = 1 - 10**-10
        #         design[:, idx] = np.array(v_design * len(param.sequence),
        #                                   dtype=np.int)
        #     else:
        #         raise ValueError("Hyperparameter not supported in LHD")

        configs = []
        for vector in design:
            conf = DenseConfiguration.from_array(cs, vector)
            # conf = deactivate_inactive_hyperparameters(configuration=None,
            #                                            configuration_space=cs,
            #                                            vector=vector)
            conf.origin = origin
            configs.append(conf)

        return configs



# License: 3-clause BSD
# Copyright (c) 2016-2018, Ml4AAD Group (http://www.ml4aad.org/)

import typing
from scipy.stats.qmc import Sobol
import numpy as np

from xbbo.configspace.space import DenseConfiguration
from xbbo.initial_design.base import InitialDesign
from xbbo.utils.constants import MAXINT


class SobolDesign(InitialDesign):
    """ Sobol sequence design with a scrambled Sobol sequence.

    See https://scipy.github.io/devdocs/reference/generated/scipy.stats.qmc.Sobol.html for further information

    Attributes
    ----------
    configs : typing.List[Configuration]
        List of configurations to be evaluated
        Don't pass configs to the constructor;
        otherwise factorial design is overwritten
    """
    def __init__(self, cs, rng: np.random.RandomState, ta_run_limit: typing.Optional[int] = None, n_configs_x_params: typing.Optional[int] = 10, init_budget: typing.Optional[int] = None, max_config_fracs: float = 0.25, **kwargs) -> None:
        super().__init__(cs, rng, ta_run_limit, n_configs_x_params, init_budget, max_config_fracs, **kwargs)
        if self.init_budget:
            self.init_budget = 2**int(np.log2(self.init_budget))

    def _select_configurations(self, num=None) -> typing.List[DenseConfiguration]:
        """Selects a single configuration to run

        Returns
        -------
        config: Configuration
            initial incumbent configuration
        """
        design_num = num if num else self.init_budget
        sobol_gen = Sobol(d=self.dim,
                          scramble=True,
                          seed=self.rng.randint(low=0, high=MAXINT))
        sobol = sobol_gen.random(design_num)
        return self._transform_continuous_designs(design=sobol,
                                                  origin='Sobol',
                                                  cs=self.cs)


import typing
import numpy as np
from xbbo.surrogate.base import SurrogateModel
from xbbo.surrogate.transfer.rf_with_instances import RandomForestWithInstances
from xbbo.utils.util import get_types

class RandomForestEnsemble(SurrogateModel):
    def __init__(self, cs, all_budgets, weight_list, fusion_method, types=None, bounds=None, **kwargs):
        if types is None or bounds is None:
            types, bounds = get_types(cs)
        super().__init__(types=types, bounds=bounds,**kwargs)

        # self.s_max = s_max
        # self.eta = eta
        self.fusion = fusion_method
        self.surrogate_weight = dict()
        self.surrogate_container = dict()
        self.all_budgets = all_budgets
        self.weight_list = weight_list
        for i, budget in enumerate(all_budgets):
            # r = int(item)
            # self.surrogate_r.append(r)
            self.surrogate_weight[budget] = self.weight_list[i]
            self.surrogate_container[budget] = RandomForestWithInstances(types=types, bounds=bounds)

    def train(self, X: np.ndarray, Y: np.ndarray, r) -> 'SurrogateModel':
        """Trains the Model on X and Y.

        Parameters
        ----------
        X : np.ndarray [n_samples, n_features (config + instance features)]
            Input data points.
        Y : np.ndarray [n_samples, n_objectives]
            The corresponding target values. n_objectives must match the
            number of target names specified in the constructor.
        r : int
            Determine which surrogate in self.surrogate_container to train.

        Returns
        -------
        self : BaseModel
        """
        self.types = self._initial_types.copy()

        if len(X.shape) != 2:
            raise ValueError('Expected 2d array, got %dd array!' % len(X.shape))
        if X.shape[1] != len(self.types):
            raise ValueError('Feature mismatch: X should have %d features, but has %d' % (X.shape[1], len(self.types)))
        if X.shape[0] != Y.shape[0]:
            raise ValueError('X.shape[0] (%s) != y.shape[0] (%s)' % (X.shape[0], Y.shape[0]))

        self.n_params = X.shape[1] - self.n_feats

        # reduce dimensionality of features of larger than PCA_DIM
        if self.pca and X.shape[0] > self.pca.n_components:
            X_feats = X[:, -self.n_feats:]
            # scale features
            X_feats = self.scaler.fit_transform(X_feats)
            X_feats = np.nan_to_num(X_feats)  # if features with max == min
            # PCA
            X_feats = self.pca.fit_transform(X_feats)
            X = np.hstack((X[:, :self.n_params], X_feats))
            if hasattr(self, "types"):
                # for RF, adapt types list
                # if X_feats.shape[0] < self.pca, X_feats.shape[1] ==
                # X_feats.shape[0]
                self.types = np.array(
                    np.hstack((self.types[:self.n_params], np.zeros((X_feats.shape[1])))),
                    dtype=np.uint,
                )

        return self._train(X, Y, r)

    def _train(self, X: np.ndarray, y: np.ndarray, r):
        self.surrogate_container[r].train(X, y)

    def _predict(self, X: np.ndarray, cov_return_type='diagonal_cov'):
        if len(X.shape) != 2:
            raise ValueError(
                'Expected 2d array, got %dd array!' % len(X.shape))
        if X.shape[1] != self.types.shape[0]:
            raise ValueError('Rows in X should have %d entries but have %d!' %
                             (self.types.shape[0], X.shape[1]))
        if self.fusion == 'idp':
            means, vars = np.zeros((X.shape[0], 1)), np.zeros((X.shape[0], 1))
            for r in self.all_budgets:
                mean, var = self.surrogate_container[r].predict(X)
                means += self.surrogate_weight[r] * mean
                vars += self.surrogate_weight[r] * self.surrogate_weight[r] * var
            return means.reshape((-1, 1)), vars.reshape((-1, 1))
        else:
            raise ValueError('Undefined Fusion Method: %s!' % self.fusion)



'''
array:cat hp use idx [0,1,2...]
array_OH:cat hp use one-hot
array_bin:cat hp use idx and then normalize[0, 1], all var in [0,1]
ref:https://github.com/ltiao/bore/blob/c137ed5f7a2d859b0053b9c20e7ed1b744186ea8/bore/plugins/hpbandster/types.py
'''
from typing import List
import ConfigSpace as CS
import ConfigSpace.hyperparameters as CSH
import numpy as np
from scipy.optimize import Bounds
from ConfigSpace.conditions import ConditionComponent
from ConfigSpace.util import deactivate_inactive_hyperparameters as _deactivate_inactive_hyperparameters
# from xbbo.configspace.warp import WARP_DICT, UNWARP_DICT
from ConfigSpace.util import get_one_exchange_neighbourhood as _get_one_exchange_neighbourhood

class Bin():
    '''
    normalize to [0, 1]
    '''
    def __init__(self, src, trg, sizes) -> None:
        self.src = src
        self.trg = trg
        self.sizes = sizes
        self.bins_width = 1 / self.sizes
        self.bins = np.array([np.arange(start=0, stop=1, step=b) for b in self.bins_width])

    def convert(self, array_dense, array_sparse):
        for i in range(len(self.src)):
            array_sparse[self.src[i]] = np.searchsorted(self.bins[i], array_dense[self.trg[i]],side="right") - 1
        # array_sparse[self.src] = np.round(array_dense[self.trg]*(self.sizes-1))
        return array_sparse
    def invconvert(self, array_dense, array_sparse):
        array_dense[self.trg] = array_sparse[self.src] / self.sizes #+ self.bins_width/2
        return array_dense
    def get_bounds(self,):
        return np.zeros(len(self.trg)), np.ones(len(self.trg))

class Round():
    def __init__(self, src, trg, sizes) -> None:
        self.src = src
        self.trg = trg
        self.sizes = sizes
    def convert(self, array_dense, array_sparse):
        array_sparse[self.src] = np.round(array_dense[self.trg])
        return array_sparse
    def invconvert(self, array_dense, array_sparse):
        array_dense[self.trg] = array_sparse[self.src]
        return array_dense
    def get_bounds(self,):
        return np.zeros(len(self.trg)), np.array((self.sizes))-1

class OneHot():
    def __init__(self, src, trg, sizes) -> None:
        self.src = src
        self.trg = trg
        self.sizes = sizes
        self.cats = list(zip(self.src.tolist(),self.trg.tolist(),self.sizes.tolist()))
    def convert(self, array_dense, array_sparse):
        for src_ind, trg_ind, size in self.cats:
            tmp = array_dense[trg_ind:trg_ind + size]
            # ind_max = np.argmax(tmp) if tmp.any() else np.nan
            ind_max = np.argmax(tmp)
            array_sparse[src_ind] = ind_max
        return array_sparse
    def invconvert(self,array_dense, array_sparse):
        choice = array_sparse[self.src] # conditional=>nan
        cat_trg_offset = np.uintp(choice)
        idx = np.where(~np.isnan(choice))[0]
        array_dense[(self.trg + cat_trg_offset)[idx]] = 1
        return array_dense
    # def get_bounds(self,):
    #     return np.zeros(self.trg), np.ones(self.trg)

class Num():
    def __init__(self, src, trg) -> None:
        self.src = src
        self.trg = trg
    def convert(self, array_dense, array_sparse):
        array_sparse[self.src] = array_dense[self.trg]
        return array_sparse
    def invconvert(self, array_dense, array_sparse):
        array_dense[self.trg] = array_sparse[self.src]
        return array_dense
    def get_bounds(self,):
        return np.zeros(len(self.trg)), np.ones(len(self.trg))

class Const():
    def __init__(self, src, values) -> None:
        self.src = src
        # self.trg = trg
        self.values = np.array(values)
    def convert(self, array_dense, array_sparse):
        array_sparse[self.src] = self.values
        return array_sparse
    def invconvert(self, array_dense, array_sparse):
        # array_dense[self.trg] = array_sparse[self.src]
        # array_dense = np.delete(array_sparse, self.src, axis=-1)
        return array_dense
    def get_bounds(self,):
        return None


class DenseConfigurationSpace(CS.ConfigurationSpace):
    def __init__(self,other: CS.ConfigurationSpace, encoding_cat, encoding_ord, *args, **kwargs):

        super().__init__(*args, **kwargs)
        # deep-copy only the hyperparameters. conditions, clauses, seed,
        # and other metadata ignored
        self.random = other.random
        hps = other.get_hyperparameters()
        self.add_hyperparameters(hps)
        # self._dim = len(hps)
        self.encoding_cat = encoding_cat
        self.encoding_ord = encoding_ord
        self._creat_mappings()
        self._create_type()

    def __len__(self):
        return self.size_sparse

    def get_dimensions(self, sparse=False):
        return self.size_dense if not sparse else self.size_sparse - self.const_hp_num

    def sample_configuration(self, size=1):
        if size == 0:
            return []

        config = super().sample_configuration(size=size)

        configs = config if size > 1 else [config]


        return [DenseConfiguration(self, values=config.get_dictionary()) for config in configs]

    def get_bounds(self):
        dim = self.get_dimensions()
        lower = np.zeros(dim)
        upper = np.ones(dim)
        for v in self.map.values():
            if isinstance(v, OneHot):
                continue
            bounds = v.get_bounds()
            if bounds is None:
                continue
            l, u = bounds
            lower[v.trg] = l
            upper[v.trg] = u
        return Bounds(lower, upper)

    def _create_type(self):
        types = []#[0] * len(self.get_hyperparameters())
        bounds = []#[(np.nan, np.nan)] * len(types)

        for i, param in enumerate(self.get_hyperparameters()):
            parents = self.get_parents_of(param.name)
            if len(parents) == 0:
                can_be_inactive = False
            else:
                can_be_inactive = True

            if isinstance(param, (CSH.CategoricalHyperparameter)):
                n_cats = len(param.choices)
                if can_be_inactive:
                    n_cats = len(param.choices) + 1
                types.append(n_cats)
                bounds.append((int(n_cats), np.nan))

            elif isinstance(param, (CSH.OrdinalHyperparameter)):
                n_cats = len(param.sequence)
                types.append(0)
                if can_be_inactive:
                    bounds.append((0, int(n_cats)))
                else:
                    bounds.append((0, int(n_cats) - 1))

            elif isinstance(param, CSH.Constant):
                # for constants we simply set types to 0 which makes it a numerical
                # parameter
                pass
                # if can_be_inactive:
                #     bounds.append((2, np.nan))
                #     types.append(2)
                # else:
                #     bounds.append((0, np.nan))
                #     types.append(0)
                # and we leave the bounds to be 0 for now
            elif isinstance(param, CSH.UniformFloatHyperparameter):
                # Are sampled on the unit hypercube thus the bounds
                # are always 0.0, 1.0 
                types.append(0)
                if can_be_inactive:
                    bounds.append((-1.0, 1.0))
                else:
                    bounds.append((0, 1.0))
            elif isinstance(param, CSH.UniformIntegerHyperparameter):
                types.append(0)
                if can_be_inactive:
                    bounds.append((-1.0, 1.0))
                else:
                    bounds.append((0, 1.0))
            elif not isinstance(param, (CSH.UniformFloatHyperparameter,
                                        CSH.UniformIntegerHyperparameter,
                                        CSH.OrdinalHyperparameter,
                                        CSH.CategoricalHyperparameter)):
                raise TypeError("Unknown hyperparameter type %s" % type(param))


        types = np.array(types, dtype=np.uint)
        bounds = np.array(bounds, dtype=object)
        self._types = types
        self._bounds = bounds

    def _creat_mappings(self):

        nums = []
        bin_cats = []
        oh_cats = []
        round_cats = []
        consts = []
        src_ind = trg_ind = 0
        for src_ind, hp in enumerate(self.get_hyperparameters()):
            if isinstance(hp, CS.CategoricalHyperparameter):
                cat_size = hp.num_choices
                if self.encoding_cat == 'one-hot':
                    oh_cats.append((src_ind, trg_ind, cat_size))
                    trg_ind += cat_size
                elif self.encoding_cat == 'bin':
                    bin_cats.append((src_ind, trg_ind, cat_size))
                    trg_ind += 1
                elif self.encoding_cat == 'round':
                    round_cats.append((src_ind, trg_ind, cat_size))
                    trg_ind += 1
                else:
                    raise ValueError()


            elif isinstance(hp, (CSH.OrdinalHyperparameter)):
                cat_size = len(hp.sequence)
                if self.encoding_cat == 'one-hot':
                    oh_cats.append((src_ind, trg_ind, cat_size))
                    trg_ind += cat_size
                elif self.encoding_cat == 'bin':
                    bin_cats.append((src_ind, trg_ind, cat_size))
                    trg_ind += 1
                elif self.encoding_cat == 'round':
                    round_cats.append((src_ind, trg_ind, cat_size))
                    trg_ind += 1
                else:
                    raise ValueError()
            elif isinstance(hp, (CS.UniformIntegerHyperparameter,
                                 CS.UniformFloatHyperparameter)):
                nums.append((src_ind, trg_ind))
                trg_ind += 1
            elif isinstance(hp, (CS.Constant)):
                value = hp.value
                if not isinstance(value, (float, int)):
                    value = 0.
                consts.append((src_ind, float(value)))
            else:
                raise NotImplementedError(
                    "Only hyperparameters of types "
                    "`CategoricalHyperparameter`, "
                    "`UniformIntegerHyperparameter`, "
                    "`UniformFloatHyperparameter`,"
                    "`OrdinalHyperparameter` are supported!")

        size_sparse = src_ind + 1
        size_dense = trg_ind
        self.map = {}
        if nums:
        # num_src, num_trg = list(map(np.uintp, zip(
        #     *nums))) if nums else ([], [])
            self.map['num'] = Num(*list(map(np.uintp, zip(
            *nums))) if nums else ([], []))

        # if cats:
        # cat_src, cat_trg, cat_sizes = \
                

        # self.nums = nums
        # self.cats = cats
        if oh_cats:
            self.map['one-hot'] = OneHot(*list(map(np.uintp, zip(*oh_cats))))
        # else:
        #     self.onehot = None
        if bin_cats:
            self.map['bin'] = Bin(*list(map(np.uintp, zip(*bin_cats))))
        # else:
        #     self.bin = None
        if round_cats:
            # raise NotImplementedError
            self.map['round'] = Round(*list(map(np.uintp, zip(*round_cats))))
        if consts:
            # raise NotImplementedError
            self.map['const'] = Const(*(zip(*consts)) if consts else ([],[]))
        # else:
        #     self.round = None
        self.size_sparse = size_sparse
        self.size_dense = size_dense
        self.const_hp_num = len(consts)

class DenseConfiguration(CS.Configuration):

    def __init__(self, configuration_space: DenseConfigurationSpace, *args,
                 **kwargs):
        '''
        values:dict
        vector:inner representation
        '''
        assert isinstance(configuration_space, DenseConfigurationSpace)
        # if 'vector' in kwargs and configuration_space.cats:
        #     kwargs['vector'][configuration_space.cat_src] = np.round(
        #         kwargs['vector'][configuration_space.cat_src])
        super(DenseConfiguration, self).__init__(configuration_space, *args,
                                                 **kwargs)

    @classmethod
    def from_dict(cls, configuration_space, dictionary):
        return cls(configuration_space=configuration_space, values=dictionary)

    @classmethod
    def from_array(cls,
                         configuration_space,
                         array_dense,
                         dtype="float64",**kwargs):
        '''
        扩展的opt给出一个suggest vector（dense array）
        需要转换成sparse array(*) => dict => unwarp dict
        '''
        cs = configuration_space
        # initialize output array
        array_sparse = np.zeros(cs.size_sparse, dtype=dtype)
        for v in cs.map.values():
            array_sparse = v.convert(array_dense, array_sparse)
        # process numerical hyperparameters
        # if cs.num:
        #     array_sparse = cs.num.convert(array_dense, array_sparse)
        # # process categorical hyperparameters
        # if cs.round:
        #     array_sparse = cs.round.convert(array_dense, array_sparse)
        # if cs.bin:
        #     array_sparse = cs.bin.convert(array_dense, array_sparse)
        # if cs.onehot:
        #     array_sparse = cs.onehot.convert(array_dense, array_sparse)
        


        return cls(configuration_space=configuration_space,
                   vector=array_sparse)

    @staticmethod
    def array_to_dict(cs, array_dense):
        # cs = space.configuration_space
        config = DenseConfiguration.from_array(
            cs, array_dense=array_dense)  # dense_array构造config类
        dict = config.get_dictionary()
        return dict

    @staticmethod
    def dict_to_array(cs, dict):
        config = DenseConfiguration(cs, values=dict)  # dict构造config类
        array_dense = config.get_array(sparse=False)  # ->4
        return array_dense  #[cs.rerangeIDX]

    def get_array(self, sparse=True, dtype="float64"):

        cs = self.configuration_space
        array_sparse = super(DenseConfiguration, self).get_array()
        if sparse:
            return array_sparse
        # initialize output array
        # TODO(LT): specify `dtype` flexibly
        array_dense = np.zeros(cs.size_dense, dtype=dtype)

        # process numerical hyperparameters
        # if cs.nums:
        #     array_dense[cs.num_trg] = array_sparse[cs.num_src]

        # # process categorical hyperparameters
        # if cs.cats:
        #     choice = array_sparse[cs.cat_src] # conditional=>nan
        #     cat_trg_offset = np.uintp(choice)
        #     array_dense[cs.cat_trg + cat_trg_offset] = 1 if choice else choice
        # # process numerical hyperparameters
        # if cs.num:
        #     array_dense = cs.num.invconvert(array_dense, array_sparse)
        # # process categorical hyperparameters
        # if cs.round:
        #     array_dense = cs.round.invconvert(array_dense, array_sparse)
        # if cs.bin:
        #     array_dense = cs.bin.invconvert(array_dense, array_sparse)
        # if cs.onehot:
        #     array_dense = cs.onehot.invconvert(array_dense, array_sparse)
        for v in cs.map.values():
            array_dense = v.invconvert(array_dense, array_sparse)
        return array_dense
 


def convert_denseConfigurations_to_array(
        configs: List[DenseConfiguration], sparse=True) -> np.ndarray:
    """Impute inactive hyperparameters in configurations with their default.

    Necessary to apply an EPM to the data.

    Parameters
    ----------
    configs : List[Configuration]
        List of configuration objects.

    Returns
    -------
    np.ndarray
        Array with configuration hyperparameters. Inactive values are imputed
        with their default value.
    """
    configs_array = np.array([config.get_array(sparse=sparse) for config in configs],
                             dtype=np.float64)
    # configuration_space = configs[0].configuration_space
    return configs_array
    # return impute_default_values(configuration_space, configs_array)

def impute_default_values(configuration_space: DenseConfigurationSpace,
                          configs_array: np.ndarray) -> np.ndarray:
    """Impute inactive hyperparameters in configuration array with their default.

    Necessary to apply an EPM to the data.

    Parameters
    ----------
    configuration_space : ConfigurationSpace
    
    configs_array : np.ndarray
        Array of configurations.

    Returns
    -------
    np.ndarray
        Array with configuration hyperparameters. Inactive values are imputed
        with their default value.
    """
    for hp in configuration_space.get_hyperparameters():
        default = hp.normalized_default_value
        idx = configuration_space.get_idx_by_hyperparameter_name(hp.name)
        nonfinite_mask = ~np.isfinite(configs_array[:, idx])
        configs_array[nonfinite_mask, idx] = default

    return configs_array


def deactivate_inactive_hyperparameters(*arg, **kwargs):
    configuration = _deactivate_inactive_hyperparameters(*arg, **kwargs)
    # return DenseConfiguration(configuration_space=configuration.configuration_space,values=configuration.get_dictionary())
    return convert_denseConfiguration(configuration)


def convert_denseConfiguration(conf: CS.Configuration):
    # return DenseConfiguration(configuration_space=conf.configuration_space,values=conf.get_dictionary())
    conf.__class__ = DenseConfiguration
    return conf


def get_one_exchange_neighbourhood(*args, **kwargs):
    configs = map(
        convert_denseConfiguration,
        _get_one_exchange_neighbourhood(*args,
                                        stdev=0.05,
                                        num_neighbors=8,
                                        **kwargs))
    return configs


import numpy as np

from . import problem_register
from xbbo.core.abstract_model import TestFunction

@problem_register.register('toy-problems')
class Model(TestFunction):
    _SUPPORT_FUNCTIONS = ('rosenbrock', 'rastrigin', 'indexsum', 'toyrebar',
                          'branin', 'ZDT1')

    def __init__(self, seed, **kwargs):
        # np.random.seed(cfg.GENERAL.random_seed)
        # self.dim = 30
        # assert self.dim % 2 == 0
        super().__init__(seed=seed)

        assert kwargs["func_name"] in self._SUPPORT_FUNCTIONS
        # func_name = cfg.TEST_PROBLEM.kwargs.func_name
        func_name = kwargs.get('func_name')
        if func_name == 'rosenbrock':
            self.func = Rosenbrock(
                kwargs.get("dim"))
        elif func_name == 'rastrigin':
            self.func = Rastrigin(
                kwargs.get("dim"))
        elif func_name == 'indexsum':
            self.func = IndexSum(kwargs.get("dim"))
        elif func_name == 'toyrebar':
            self.func = ToyREBAR(kwargs.get("dim"),
                                 kwargs.get("t"))
        elif func_name == 'branin':
            self.func = Branin()
        elif func_name == 'ZDT1':
            self.func = ZDT1()
        else:
            assert False
        self.noise_std = kwargs.get('noise_std', 0)
        self.api_config = self._load_api_config()

    def evaluate(self, params: dict):

        input_x = []
        for k in sorted(params.keys()):
            input_x.append(params[k])
        f = self.func(np.asarray(input_x))
        if isinstance(f, tuple):
            random_noise = self.rng.randn(len(f)) * self.noise_std + 1.
        else:
            random_noise = self.rng.randn() * self.noise_std + 1.
        # res_out = {
        #     # 'raw': f,
        #     'noise': f * random_noise,
        # }
        # res_loss = {
        #     'test': f,
        #     'val': f * random_noise,
        # }
        # return (res_out['noise'])
        return f * random_noise

    def _load_api_config(self):
        return self.func._load_api_config()


class Rastrigin():
    def __init__(self, dim):
        self.dim = dim

    def __call__(self, input_x):
        f_x = 10. * input_x.shape[0]
        for i in input_x:
            f_x += i**2 - 10 * np.cos(2 * np.pi * i)
        return f_x

    def _load_api_config(self):
        return {
            'x_{}'.format(k): {
                'type': 'float',
                'range': (-5.12, 5.12)
            }
            for k in range(self.dim)
        }


class IndexSum():
    def __init__(self, dim):
        self.dim = dim

    def __call__(self, input_x):
        f_x = 0
        for i in input_x:
            f_x += i  # cat都取第一个类最优
        return f_x

    def _load_api_config(self):
        return {
            'x_{}'.format(k): {
                'type': 'cat',
                'values': list(range(10))
            }
            for k in range(self.dim)
        }


class ToyREBAR():
    # min
    def __init__(self, dim, t=0.45):
        # self.a = 1
        # self.b = 100
        self.dim = dim
        self.t = t
        pass

    def __call__(self, input_x):
        """
        Rosenbrock function
        """
        f_x = 0
        for i in input_x:
            f_x += (i - self.t)**2

        return f_x / input_x.shape[0]

    # def _load_api_config(self):
    #     return {
    #         'x_{}'.format(k): {'type': 'cat', 'values': list(range(-5,6))} for k in range(self.dim)
    #     }

    def _load_api_config(self):
        '''
        参数是b的value
        '''
        return {
            'b_{}'.format(k): {
                'type': 'cat',
                'values': (0, 1)
            }
            for k in range(self.dim)
        }


class Rosenbrock():
    # min
    def __init__(self, dim):
        # self.a = 1
        # self.b = 100
        self.dim = dim
        pass

    def __call__(self, input_x):
        """
        Rosenbrock function
        """
        f_x = 0
        for i in range(input_x.shape[0] - 1):
            f_x += 100 * (input_x[i + 1] - input_x[i]**2)**2 + (1 -
                                                                input_x[i])**2

        return f_x

    # def _load_api_config(self):
    #     return {
    #         'x_{}'.format(k): {'type': 'cat', 'values': list(range(-5,6))} for k in range(self.dim)
    #     }

    def _load_api_config(self):
        return {
            'x_{}'.format(k): {
                'type': 'float',
                'range': (-5, 10)
            }
            for k in range(self.dim)
        }


class Branin():
    '''
    The Branin, or Branin-Hoo, function has three global minima. The recommended values of a, b, c, r, s and t are: a = 1, b = 5.1 ⁄ (4π2), c = 5 ⁄ π, r = 6, s = 10 and t = 1 ⁄ (8π).
    https://www.sfu.ca/~ssurjano/branin.html
    global minimum:
    f(x*) = 0.397887
    x* = (-pi, 12.275)、(pi, 2.275)、(9.42478, 2.475)
    '''
    def __init__(self, ):
        self.a = 1
        self.b = 5.1 / (4 * np.pi**2)
        self.c = 5 / np.pi
        self.r = 6
        self.s = 10
        self.t = 1 / (8 * np.pi)

    def __call__(self, input_x):
        x1, x2 = input_x
        term1 = self.a * (x2 - self.b * x1**2 + self.c * x1 - self.r)**2
        term2 = self.s * (1 - self.t) * np.cos(x1)

        y = term1 + term2 + self.s
        return y

    def _load_api_config(self):
        return {
            "x1": {
                "type": "float",
                "range": [-5, 10]
            },
            "x2": {
                "type": "float",
                "range": [0, 15]
            }
        }


class ZDT1():

    def __init__(self, ):
        self.n = 30

    def __call__(self, input_x):
        sigma = sum(input_x[1:])
        g = 1 + sigma * 9 / (self.n - 1)
        h = 1 - (input_x[0] / g)**0.5
        return input_x[0], g*h

    def _load_api_config(self):
        return {
            "x1": {
                "type": "float",
                "range": [0, 1]
            },
            "x2": {
                "type": "float",
                "range": [0, 1]
            }
        }



import logging
import typing
import numpy as np

from xbbo.acquisition_function.acq_optimizer import InterleavedLocalAndRandomSearch, LocalSearch, RandomScipyOptimizer, RandomSearch, ScipyGlobalOptimizer, ScipyOptimizer

from xbbo.search_algorithm.base import AbstractOptimizer
# from xbbo.configspace.space import DenseConfiguration, DenseConfigurationSpace

# from xbbo.core import trials
# from xbbo.core.stochastic import Category, Uniform
from . import alg_register
from xbbo.core.trials import Trial, Trials
from xbbo.initial_design import ALL_avaliable_design
from xbbo.surrogate.gaussian_process import GPR_sklearn
from xbbo.acquisition_function.acq_func import EI_AcqFunc
from xbbo.surrogate.sk_prf import skRandomForestWithInstances
from xbbo.surrogate.skrf import RandomForestSurrogate

logger = logging.getLogger(__name__)


@alg_register.register('basic-bo')
class BO(AbstractOptimizer):
    def __init__(
            self,
            space,
            seed: int = 42,
            surrogate: str = 'gp',
            acq_func: str = 'ei',
            acq_opt: str = 'rs_ls',
            initial_design: str = 'sobol',
            #  min_sample=1,
            suggest_limit: int = np.inf,
            predict_x_best: bool = True,
            **kwargs):
        '''
        predict_x_best: bool
            Choose x_best for computing the acquisition function via the model instead of via the observations.
        '''
        AbstractOptimizer.__init__(self,
                                   space,
                                   encoding_cat='bin',
                                   encoding_ord='bin',
                                   seed=seed,
                                   suggest_limit=suggest_limit,
                                   **kwargs)
        # self.min_sample = min_sample
        # configs = self.space.get_hyperparameters()
        self.predict_x_best = predict_x_best
        self.dimension = self.space.get_dimensions(sparse=True)

        self.initial_design = ALL_avaliable_design[initial_design](
            self.space, self.rng, ta_run_limit=suggest_limit, **kwargs)
        self.init_budget = self.initial_design.init_budget
        self.initial_design_configs = self.initial_design.select_configurations(
        )

        self.trials = Trials(space, self.dimension)
        if surrogate == 'gp':
            self.surrogate_model = GPR_sklearn(self.space, rng=self.rng)
        elif surrogate == 'prf':
            from xbbo.surrogate.prf import RandomForestWithInstances
            self.surrogate_model = RandomForestWithInstances(self.space,
                                                             rng=self.rng)
        elif surrogate == 'rf':
            self.surrogate_model = RandomForestSurrogate(self.space,
                                                         rng=self.rng)
        elif surrogate == 'sk_prf':
            self.surrogate_model = skRandomForestWithInstances(self.space,
                                                               rng=self.rng)
        else:
            raise ValueError('surrogate {} not in {}'.format(
                surrogate, ['gp', 'rf', 'prf', 'sk_prf']))

        if acq_func == 'ei':
            self.acquisition_func = EI_AcqFunc(self.surrogate_model, self.rng)
        # elif acq_func == 'rf':
        #     self.acquisition_func = None
        else:
            raise ValueError('acq_func {} not in {}'.format(acq_func, ['ei']))
        if acq_opt == 'ls':
            self.acq_maximizer = LocalSearch(self.acquisition_func, self.space,
                                             self.rng)
        elif acq_opt == 'rs':
            self.acq_maximizer = RandomSearch(self.acquisition_func,
                                              self.space, self.rng)
        elif acq_opt == 'rs_ls':
            self.acq_maximizer = InterleavedLocalAndRandomSearch(
                self.acquisition_func, self.space, self.rng)
        elif acq_opt == 'scipy':
            self.acq_maximizer = ScipyOptimizer(self.acquisition_func,
                                                self.space, self.rng)
        elif acq_opt == 'scipy_global':
            self.acq_maximizer = ScipyGlobalOptimizer(self.acquisition_func,
                                                      self.space, self.rng)
        elif acq_opt == 'r_scipy':
            self.acq_maximizer = RandomScipyOptimizer(self.acquisition_func,
                                                      self.space, self.rng)
        else:
            raise ValueError('acq_opt {} not in {}'.format(
                acq_opt,
                ['ls', 'rs', 'rs_ls', 'scipy', 'scipy_global', 'r_scipy']))
        logger.info(
            "Execute Bayesian optimization...\n [Using ({})surrogate, ({})acquisition function, ({})acquisition optmizer]"
            .format(surrogate, acq_func, acq_opt))

    def _suggest(self, n_suggestions=1):
        trial_list = []
        # currently only suggest one
        if (self.trials.trials_num) < self.init_budget:
            assert self.trials.trials_num % n_suggestions == 0
            configs = self.initial_design_configs[
                int(n_suggestions *
                    self.trials.trials_num):int(n_suggestions *
                                                (self.trials.trials_num + 1))]
            for config in configs:
                trial_list.append(
                    Trial(configuration=config,
                          config_dict=config.get_dictionary(),
                          array=config.get_array()))
        else:
            self.surrogate_model.train(
                np.asarray(self.trials.get_array()),
                np.asarray(self.trials.get_history()[0]))
            configs = []
            _, best_val = self._get_x_best(self.predict_x_best)
            self.acquisition_func.update(surrogate_model=self.surrogate_model,
                                         y_best=best_val)
            configs = self.acq_maximizer.maximize(self.trials,
                                                  1000,
                                                  drop_self_duplicate=True)
            _idx = 0
            for n in range(n_suggestions):
                while _idx < len(configs):  # remove history suggest
                    if not self.trials.is_contain(configs[_idx]):
                        config = configs[_idx]
                        configs.append(config)
                        trial_list.append(
                            Trial(configuration=config,
                                  config_dict=config.get_dictionary(),
                                  array=config.get_array()))
                        _idx += 1

                        break
                    _idx += 1
                else:
                    assert False, "no more configs can be suggest"

        return trial_list

    def _observe(self, trial_list):
        for trial in trial_list:
            self.trials.add_a_trial(trial)

    def _get_x_best(self, predict: bool) -> typing.Tuple[float, np.ndarray]:
        """Get value, configuration, and array representation of the "best" configuration.

        The definition of best varies depending on the argument ``predict``. If set to ``True``,
        this function will return the stats of the best configuration as predicted by the model,
        otherwise it will return the stats for the best observed configuration.

        Parameters
        ----------
        predict : bool
            Whether to use the predicted or observed best.

        Returns
        -------
        float
        np.ndarry
        Configuration
        """
        if predict:
            X = self.trials.get_array()
            costs = list(
                map(
                    lambda x: (
                        self.surrogate_model.predict(x.reshape((1, -1)))[0][0],
                        x,
                    ),
                    X,
                ))
            costs = sorted(costs, key=lambda t: t[0])
            x_best_array = costs[0][1]
            best_observation = costs[0][0]
            # won't need log(y) if EPM was already trained on log(y)
        else:
            best_idx = self.trials.best_id
            x_best_array = self.trials.get_array()[best_idx]
            best_observation = self.trials.best_observe_value

        return x_best_array, best_observation


opt_class = BO


from typing import Iterable
import numpy as np
import ConfigSpace as CS
from xbbo.configspace.space import DenseConfiguration
from xbbo.utils.constants import Key


class Trial:
    def __init__(self,
                 configuration,
                 config_dict,
                 observe_value=None,
                 array=None,
                #  sparse_array=None,
                 time=None,
                 origin: str = '',
                 marker=None,
                 info: dict=None,
                 **kwargs) -> None:
        self.config_dict = config_dict
        self.array = array
        # self.sparse_array = sparse_array
        self.configuration = configuration
        self.observe_value = observe_value
        self.time = time
        self.origin = origin
        self.marker = marker
        self.info = info if info else {}
        for k in kwargs:
            setattr(self, k, kwargs[k])

    def add_observe_value(self, observe_value=None, obs_info=None):
        if obs_info is None:
            obs_info = {}
        self.info.update(obs_info)
        self.observe_value = observe_value


class Trials:
    def __init__(self, cs, dim):
        self.cs = cs
        # self._non_const_idx = []
        # for i, hp in enumerate(cs.get_hyperparameters()):
        #     if not isinstance(hp, CS.Constant):
        #         self._non_const_idx.append(i)
        # self._non_const_idx = np.array(self._non_const_idx)
        self._his_hash_configs_set = set()
        self._his_configs_set = set()
        self._his_configs = []
        self._his_array = None #np.empty((0, dim))
        self.dim = dim
        self._his_observe_value = []
        self._his_configs_dict = []
        self.best_observe_value = np.inf
        self.best_id = None
        self.trials_num = 0
        self.markers = []
        self.infos = []
        # self.run_id = 0
        # self.run_history = {}
        self.traj_history = []
        # self.use_dense = use_dense

    def add_a_trial(self, trial: Trial, permit_duplicate=True):
        hash_config = str(trial.configuration)+str(trial.info.get(Key.BUDGET, 'max_budget'))
        if not permit_duplicate:
            assert hash_config not in self._his_hash_configs_set
        self.infos.append(trial.info)
        self._his_hash_configs_set.add(hash_config)
        self._his_configs_set.add(trial.configuration)
        self._his_configs.append(trial.configuration)
        self.traj_history.append(trial)
        self._his_configs_dict.append(trial.config_dict)
        self._his_observe_value.append(trial.observe_value)
        self.markers.append(trial.marker)
        # if self.use_dense:
        #     assert trial.dense_array is not None
        # else:
        #     assert trial.sparse_array is not None
        if trial.array is not None:
            # trial.array = np.atleast_2d(trial.array)
            # if self.dim != trial.array.shape[-1]:
            #     trial.array = (trial.array[...,self._non_const_idx])
            if self._his_array is None:
                self._his_array = np.empty((0,trial.array.shape[-1]))
            self._his_array = np.vstack(
                [self._his_array, trial.array])
        # if trial.sparse_array is not None:
        #     self._his_sparse_array = np.vstack(
        #         [self._his_sparse_array, trial.sparse_array])
        obs = sum(trial.observe_value) if isinstance(trial.observe_value, Iterable) else trial.observe_value
        if self.best_observe_value > obs:
            self.best_observe_value = obs
            self.best_id = self.trials_num
        self.trials_num += 1

    def get_array(self):
        if  self._his_array is not None and len(self._his_array) == self.trials_num:
            return self._his_array
        if self.trials_num == 0:
            return None
        self._his_array = [
            config.get_array(sparse=False) for config in self._his_configs
        ]
        return self._his_array

    # def get_sparse_array(self):
    #     if len(self._his_sparse_array) == self.trials_num:
    #         return self._his_sparse_array
    #     self._his_sparse_array = [
    #         config.get_sparse_array() for config in self._his_configs
    #     ]
    #     return self._his_sparse_array

    def add_trials(self, trials):
        for trial in trials._traj_history:
            self.add_a_trial(trial)

    def is_contain(self, config: DenseConfiguration) -> bool:
        return config in self._his_hash_configs_set

    def is_empty(self, ):
        return self.trials_num == 0

    def get_all_configs(self, ):
        return self._his_configs

    def get_best(self):
        return self.best_observe_value, self._his_configs_dict[self.best_id]

    def get_history(self):
        return self._his_observe_value, self._his_configs_dict
    
    # def get_array(self):
    #     if self.use_dense:
    #         return self.get_dense_array()
    #     else:
    #         return self.get_sparse_array()

    # def visualize(self, ax=None):
    #     if ax is None:
    #         _, ax = plt.subplots(111)
    #     ax.plot(np.minimum.acumalate(self._his_observe_value))


import numpy as np
from openbox.surrogate.base.base_model import AbstractModel
from openbox.surrogate.base.build_gp import create_gp_model


class GaussianProcessEnsemble(AbstractModel):
    def __init__(self, config_space, types: np.ndarray,
                 bounds: np.ndarray, s_max, eta, weight_list, fusion_method, rng, **kwargs):
        super().__init__(types=types, bounds=bounds, **kwargs)

        self.s_max = s_max
        self.eta = eta
        self.fusion = fusion_method
        self.surrogate_weight = dict()
        self.surrogate_container = dict()
        self.surrogate_r = list()
        self.weight_list = weight_list
        self.rng = rng
        for index, item in enumerate(np.logspace(0, self.s_max, self.s_max + 1, base=self.eta)):
            r = int(item)
            self.surrogate_r.append(r)
            self.surrogate_weight[r] = self.weight_list[index]
            self.surrogate_container[r] = create_gp_model('gp', config_space, types, bounds, self.rng)

    def train(self, X: np.ndarray, Y: np.ndarray, r) -> 'AbstractModel':
        """Trains the Model on X and Y.

        Parameters
        ----------
        X : np.ndarray [n_samples, n_features (config + instance features)]
            Input data points.
        Y : np.ndarray [n_samples, n_objectives]
            The corresponding target values. n_objectives must match the
            number of target names specified in the constructor.
        r : int
            Determine which surrogate in self.surrogate_container to train.

        Returns
        -------
        self : AbstractModel
        """
        self.types = self._initial_types.copy()

        if len(X.shape) != 2:
            raise ValueError('Expected 2d array, got %dd array!' % len(X.shape))
        if X.shape[1] != len(self.types):
            raise ValueError('Feature mismatch: X should have %d features, but has %d' % (X.shape[1], len(self.types)))
        if X.shape[0] != Y.shape[0]:
            raise ValueError('X.shape[0] (%s) != y.shape[0] (%s)' % (X.shape[0], Y.shape[0]))

        self.n_params = X.shape[1] - self.n_feats

        # reduce dimensionality of features of larger than PCA_DIM
        if self.pca and X.shape[0] > self.pca.n_components:
            X_feats = X[:, -self.n_feats:]
            # scale features
            X_feats = self.scaler.fit_transform(X_feats)
            X_feats = np.nan_to_num(X_feats)  # if features with max == min
            # PCA
            X_feats = self.pca.fit_transform(X_feats)
            X = np.hstack((X[:, :self.n_params], X_feats))
            if hasattr(self, "types"):
                # for RF, adapt types list
                # if X_feats.shape[0] < self.pca, X_feats.shape[1] ==
                # X_feats.shape[0]
                self.types = np.array(
                    np.hstack((self.types[:self.n_params], np.zeros((X_feats.shape[1])))),
                    dtype=np.uint,
                )

        return self._train(X, Y, r)

    def _train(self, X: np.ndarray, y: np.ndarray, r):
        self.surrogate_container[r].train(X, y)

    def _predict(self, X: np.ndarray):
        if len(X.shape) != 2:
            raise ValueError(
                'Expected 2d array, got %dd array!' % len(X.shape))
        if X.shape[1] != self.types.shape[0]:
            raise ValueError('Rows in X should have %d entries but have %d!' %
                             (self.types.shape[0], X.shape[1]))
        if self.fusion == 'idp':
            means, vars = np.zeros((X.shape[0], 1)), np.zeros((X.shape[0], 1))
            for r in self.surrogate_r:
                mean, var = self.surrogate_container[r].predict(X)
                means += self.surrogate_weight[r] * mean
                vars += self.surrogate_weight[r] * self.surrogate_weight[r] * var
            return means.reshape((-1, 1)), vars.reshape((-1, 1))
        else:
            raise ValueError('Undefined Fusion Method: %s!' % self.fusion)



import logging
import traceback
import numpy as np
import statsmodels.api as sm
import scipy.stats as sps
import ConfigSpace

from xbbo.search_algorithm.base import AbstractOptimizer
from xbbo.configspace.space import DenseConfiguration, DenseConfigurationSpace, deactivate_inactive_hyperparameters
from xbbo.core.trials import Trial, Trials
from xbbo.utils.constants import MAXINT
from . import alg_register
from xbbo.initial_design import ALL_avaliable_design

logger = logging.getLogger(__name__)


@alg_register.register('tpe')
class TPE(AbstractOptimizer):
    '''
    reference: https://github.com/thomas-young-2013/open-box/blob/master/openbox/core/tpe_advisor.py
    '''
    def __init__(
            self,
            space: DenseConfigurationSpace,
            seed: int = 42,
            #  bandwidth=1,
            gamma=0.15,
            initial_design: str = 'sobol',
            suggest_limit: int = np.inf,
            candidates_num=64,
            min_bandwidth=1e-3,
            bandwidth_factor=3,
            min_points_in_model=None,
            random_fraction=1 / 3,
            **kwargs):
        AbstractOptimizer.__init__(self,
                                   space,
                                   encoding_cat='round',
                                   encoding_ord='round',
                                   seed=seed,
                                   suggest_limit=suggest_limit,
                                   **kwargs)
        self.min_bandwidth = min_bandwidth
        self.bw_factor = bandwidth_factor

        self.dimension = self.space.get_dimensions()
        self.initial_design = ALL_avaliable_design[initial_design](
            self.space, self.rng, ta_run_limit=suggest_limit, **kwargs)
        self.init_budget = self.initial_design.init_budget
        self.initial_design_configs = self.initial_design.select_configurations(
        )

        self.trials = Trials(space,dim=self.dimension)
        self.gamma = gamma
        self.candidates_num = candidates_num
        self.min_points_in_model = min_points_in_model

        dim = self.space.get_dimensions(sparse=True)
        hps = self.space.get_hyperparameters()

        if min_points_in_model is None:
            self.min_points_in_model = dim + 1

        if self.min_points_in_model < dim + 1:
            self.min_points_in_model = dim + 1

        self.random_fraction = random_fraction
        self.kde_models = dict()

        self.kde_vartypes = ""
        self.vartypes = []

        for h in hps:
            if hasattr(h, 'choices'):
                self.kde_vartypes += 'u'
                self.vartypes += [len(h.choices)]
            else:
                self.kde_vartypes += 'c'
                self.vartypes += [0]

        self.vartypes = np.array(self.vartypes, dtype=int)

    def _suggest(self, n_suggestions=1):
        trial_list = []
        if (self.trials.trials_num) < self.init_budget:
            assert self.trials.trials_num % n_suggestions == 0
            configs = self.initial_design_configs[
                int(n_suggestions *
                    self.trials.trials_num):int(n_suggestions *
                                                (self.trials.trials_num + 1))]
            for config in configs:
                trial_list.append(
                    Trial(configuration=config,
                          config_dict=config.get_dictionary(),
                          array=config.get_array()))
        else:
            self._fit_kde_models()
            if len(self.kde_models.keys()
                   ) == 0 or self.rng.rand() < self.random_fraction:
                configs = self._sample_nonduplicate_config(n_suggestions)
                for config in configs:
                    trial_list.append(
                        Trial(configuration=config,
                              config_dict=config.get_dictionary(),
                              array=config.get_array()))
            else:
                for n in range(n_suggestions):
                    try:
                        best = np.inf
                        best_vector = None
                        l = self.kde_models['good'].pdf
                        g = self.kde_models['bad'].pdf

                        minimize_me = lambda x: max(1e-32, g(x)) / max(l(x), 1e-32)

                        kde_good = self.kde_models['good']
                        kde_bad = self.kde_models['bad']

                        for i in range(self.candidates_num):
                            idx = self.rng.randint(0, len(kde_good.data))
                            datum = kde_good.data[idx]
                            vector = []

                            for m, bw, t in zip(datum, kde_good.bw, self.vartypes):

                                bw = max(bw, self.min_bandwidth)
                                if t == 0:
                                    bw = self.bw_factor * bw
                                    try:
                                        vector.append(
                                            sps.truncnorm.rvs(-m / bw,
                                                            (1 - m) / bw,
                                                            loc=m,
                                                            scale=bw))
                                    except:
                                        logger.warning(
                                            "Truncated Normal failed for:\ndatum=%s\nbandwidth=%s\nfor entry with value %s"
                                            % (datum, kde_good.bw, m))
                                        logger.warning("data in the KDE:\n%s" %
                                                    kde_good.data)
                                else:

                                    if self.rng.rand() < (1 - bw):
                                        vector.append(int(m))
                                    else:
                                        vector.append(self.rng.randint(t))
                            val = minimize_me(vector)

                            if not np.isfinite(val):
                                logger.warning(
                                    'sampled vector: %s has EI value %s' %
                                    (vector, val))
                                logger.warning("data in the KDEs:\n%s\n%s" %
                                            (kde_good.data, kde_bad.data))
                                logger.warning("bandwidth of the KDEs:\n%s\n%s" %
                                            (kde_good.bw, kde_bad.bw))
                                logger.warning("l(x) = %s" % (l(vector)))
                                logger.warning("g(x) = %s" % (g(vector)))

                                # right now, this happens because a KDE does not contain all values for a categorical parameter
                                # this cannot be fixed with the statsmodels KDE, so for now, we are just going to evaluate this one
                                # if the good_kde has a finite value, i.e. there is no config with that value in the bad kde, so it shouldn't be terrible.
                                if np.isfinite(l(vector)):
                                    best_vector = vector
                                    break

                            if val < best:
                                best = val
                                best_vector = vector

                        if best_vector is None:
                            logger.debug(
                                "Sampling based optimization with %i samples failed -> using random configuration"
                                % self.candidates_num)
                            config = self._sample_nonduplicate_config()[0]
                        else:
                            logger.debug('best_vector: {}, {}, {}, {}'.format(
                                best_vector, best, l(best_vector), g(best_vector)))
                            for i, hp_value in enumerate(best_vector):
                                if isinstance(
                                        self.space.get_hyperparameter(
                                            self.space.get_hyperparameter_by_idx(
                                                i)), ConfigSpace.hyperparameters.
                                        CategoricalHyperparameter):
                                    best_vector[i] = int(np.rint(best_vector[i]))
                            config = DenseConfiguration.from_array(
                            self.space, np.asarray(best_vector))
                        try:
                            config = deactivate_inactive_hyperparameters(
                                        configuration_space=self.space,
                                        configuration=config.get_dictionary()
                                        )

                        except Exception as e:
                            logger.warning(("="*50 + "\n")*3 +\
                                    "Error converting configuration:\n%s"%config+\
                                    "\n here is a traceback:" +\
                                    traceback.format_exc())
                            raise(e)


                    except:
                        logger.warning("Sampling based optimization with %i samples failed\n %s \nUsing random configuration"%(self.num_samples, traceback.format_exc()))
                        # config = self._sample_nonduplicate_config()[0]
                        config = self.space.sample_configuration()[0]
                    trial_list.append(
                            Trial(configuration=config,
                                config_dict=config.get_dictionary(),
                                array=config.get_array()))
        return trial_list

    def _sample_nonduplicate_config(self, num_configs=1):
        configs = list()
        sample_cnt = 0
        max_sample_cnt = 1000
        while len(configs) < num_configs:
            config = self.space.sample_configuration()[0]
            sample_cnt += 1
            if (not self.trials.is_contain(config)) and config not in configs:
                configs.append(config)
                sample_cnt = 0
                continue
            if sample_cnt >= max_sample_cnt:
                logger.warning(
                    'Cannot sample non duplicate configuration after %d iterations.'
                    % max_sample_cnt)
                configs.append(config)
                sample_cnt = 0
        return configs

    def _fit_kde_models(self, ):
        train_configs = self.trials.get_array()
        if train_configs is None:
            return
        n_good = max(self.min_points_in_model,
                     int(self.gamma * self.trials.trials_num) // 100)
        # n_bad = min(max(self.min_points_in_model, ((100-self.top_n_percent)*train_configs.shape[0])//100), 10)
        n_bad = max(self.min_points_in_model,
                    int((1 - self.gamma) * self.trials.trials_num))

        # Refit KDE for the current budget
        idx = np.argsort(self.trials._his_observe_value)

        train_data_good = self.impute_conditional_data(
            train_configs[idx[:n_good]])
        train_data_bad = self.impute_conditional_data(
            train_configs[idx[n_good:n_good + n_bad]])

        if train_data_good.shape[0] <= train_data_good.shape[1]:
            return
        if train_data_bad.shape[0] <= train_data_bad.shape[1]:
            return

        bw_estimation = 'normal_reference'
        np.random.seed(self.rng.randint(MAXINT))
        bad_kde = sm.nonparametric.KDEMultivariate(data=train_data_bad,
                                                   var_type=self.kde_vartypes,
                                                   bw=bw_estimation)
        good_kde = sm.nonparametric.KDEMultivariate(data=train_data_good,
                                                    var_type=self.kde_vartypes,
                                                    bw=bw_estimation)

        bad_kde.bw = np.clip(bad_kde.bw, self.min_bandwidth, None)
        good_kde.bw = np.clip(good_kde.bw, self.min_bandwidth, None)

        self.kde_models = {'good': good_kde, 'bad': bad_kde}

    def _observe(self, trial_list):
        for trial in trial_list:
            self.trials.add_a_trial(trial, permit_duplicate=True)

    def impute_conditional_data(self, array):

        return_array = np.empty_like(array)

        for i in range(array.shape[0]):
            datum = np.copy(array[i])
            nan_indices = np.argwhere(np.isnan(datum)).ravel()

            while np.any(nan_indices):
                nan_idx = nan_indices[0]
                valid_indices = np.argwhere(np.isfinite(
                    array[:, nan_idx])).ravel()

                if len(valid_indices) > 0:
                    # pick one of them at random and overwrite all NaN values
                    row_idx = self.rng.choice(valid_indices)
                    datum[nan_indices] = array[row_idx, nan_indices]

                else:
                    # no good point in the data has this value activated, so fill it with a valid but random value
                    t = self.vartypes[nan_idx]
                    if t == 0:
                        datum[nan_idx] = self.rng.rand()
                    else:
                        datum[nan_idx] = self.rng.randint(t)

                nan_indices = np.argwhere(np.isnan(datum)).ravel()
            return_array[i, :] = datum
        return return_array


opt_class = TPE


VERY_SMALL_NUMBER = 1e-10
MAXINT = 2 ** 31 - 1
N_TREES = 10

class Key():
    REGRET_VAL = "regret_validation"
    REGRET_TEST = "regret_test"
    COST = "cost"
    BUDGET = "budget"
    EVAL_TIME = "eval_time"
    FUNC_VALUE = "function_value"